"""SAE-based steering method for comparison experiments."""

from __future__ import annotations

import gc
import json
from pathlib import Path
from typing import TYPE_CHECKING

import torch

from wisent.core.utils.config_tools.constants import DISPLAY_TOP_N_SMALL, PROGRESS_LOG_INTERVAL_10
from wisent.comparison.utils import (
    apply_steering_to_model,
    remove_steering,
    convert_to_lm_eval_format,
    generate_contrastive_pairs,
    load_model_and_tokenizer,
    load_sae,
    SAE_CONFIGS,
)

if TYPE_CHECKING:
    from wisent.core.primitives.models.wisent_model import WisentModel

__all__ = ["generate_steering_vector", "apply_steering_to_model", "remove_steering", "convert_to_lm_eval_format"]


def _get_residual_stream_activations(
    model,
    tokenizer,
    text: str,
    layer_idx: int,
    device: str,
) -> torch.Tensor:
    """
    Get residual stream activations from a specific layer.

    Args:
        model: HuggingFace model
        tokenizer: Tokenizer
        text: Input text
        layer_idx: Layer index (0-indexed)
        device: Device

    Returns:
        Tensor of shape (1, seq_len, d_model)
    """
    inputs = tokenizer(text, return_tensors="pt", truncation=True, max_length=tokenizer.model_max_length)
    inputs = {k: v.to(device) for k, v in inputs.items()}

    with torch.no_grad():
        out = model(**inputs, output_hidden_states=True, use_cache=False)

    # hidden_states is tuple: (embedding, layer0, layer1, ..., layerN)
    # layer_idx=0 -> hs[1], layer_idx=12 -> hs[13]
    hs = out.hidden_states
    return hs[layer_idx + 1]


def compute_feature_diff(
    model,
    tokenizer,
    sae,
    pairs: list[dict],
    layer_idx: int,
    device: str,
) -> torch.Tensor:
    """
    Compute feature difference between positive and negative examples in SAE space.

    feature_diff = mean(encode(h_pos)) - mean(encode(h_neg))

    Args:
        model: HuggingFace model
        tokenizer: Tokenizer
        sae: SAE object from sae_lens
        pairs: List of contrastive pairs
        layer_idx: Layer to extract activations from
        device: Device

    Returns:
        feature_diff tensor of shape [d_sae]
    """
    pos_features_list = []
    neg_features_list = []

    print(f"   Computing feature diff from {len(pairs)} pairs...")

    for i, pair in enumerate(pairs):
        prompt = pair["prompt"]
        pos_response = pair["positive_response"]["model_response"]
        neg_response = pair["negative_response"]["model_response"]

        pos_text = f"{prompt} {pos_response}"
        neg_text = f"{prompt} {neg_response}"

        # Get activations and encode through SAE
        pos_acts = _get_residual_stream_activations(model, tokenizer, pos_text, layer_idx, device)
        pos_acts = pos_acts.to(device).to(sae.W_enc.dtype)
        pos_latents = sae.encode(pos_acts)
        pos_features_list.append(pos_latents.mean(dim=1).detach())  # Mean over sequence

        neg_acts = _get_residual_stream_activations(model, tokenizer, neg_text, layer_idx, device)
        neg_acts = neg_acts.to(device).to(sae.W_enc.dtype)
        neg_latents = sae.encode(neg_acts)
        neg_features_list.append(neg_latents.mean(dim=1).detach())

        if (i + 1) % PROGRESS_LOG_INTERVAL_10 == 0:
            print(f"      Processed {i + 1}/{len(pairs)} pairs")

    # Stack and compute mean difference
    pos_features = torch.cat(pos_features_list, dim=0)  # [num_pairs, d_sae]
    neg_features = torch.cat(neg_features_list, dim=0)

    feature_diff = pos_features.mean(dim=0) - neg_features.mean(dim=0)  # [d_sae]

    print(f"   feature_diff computed, shape: {feature_diff.shape}")
    print(f"   feature_diff stats: mean={feature_diff.mean():.6f}, std={feature_diff.std():.6f}")

    return feature_diff


def compute_steering_vector_from_decoder(
    feature_diff: torch.Tensor,
    sae,
    top_k: int,
    *,
    normalize: bool = True,
) -> tuple[torch.Tensor, dict]:
    """
    Compute steering vector using SAE decoder features.

    steering_vector = sum(feature_diff[i] * W_dec[i]) for top-k features

    Args:
        feature_diff: Difference vector in SAE space [d_sae]
        sae: SAE object with W_dec decoder weights
        top_k: Number of top features to use
        normalize: Whether to normalize the final steering vector

    Returns:
        Tuple of (steering_vector [d_model], feature_info dict)
    """
    # Find top-k features by absolute difference
    abs_diff = feature_diff.abs()
    top_values, top_indices = abs_diff.topk(min(top_k, len(feature_diff)))

    print(f"   Selected top {len(top_indices)} features")
    print(f"   Top feature indices: {top_indices[:DISPLAY_TOP_N_SMALL].tolist()}...")
    print(f"   Top feature diff magnitudes: {top_values[:DISPLAY_TOP_N_SMALL].tolist()}")

    # Construct steering vector from decoder
    # W_dec shape: [d_sae, d_model]
    steering_vector = torch.zeros(sae.W_dec.shape[1], device=sae.W_dec.device, dtype=sae.W_dec.dtype)

    for feat_idx in top_indices:
        steering_vector += feature_diff[feat_idx] * sae.W_dec[feat_idx]

    if normalize:
        norm = steering_vector.norm()
        if norm > 0:
            steering_vector = steering_vector / norm
            print(f"   Normalized steering vector (original norm: {norm:.4f})")

    print(f"   steering_vector shape: {steering_vector.shape}, norm: {steering_vector.norm():.6f}")

    feature_info = {
        "top_k": top_k,
        "top_indices": top_indices.tolist(),
        "top_diff_values": [feature_diff[i].item() for i in top_indices],
    }

    return steering_vector, feature_info


def generate_steering_vector(
    task: str,
    model_name: str,
    output_path: str | Path,
    trait_label: str,
    method: str,
    device: str,
    num_pairs: int,
    layers: str,
    top_k: int,
    *,
    normalize: bool = True,
    keep_intermediate: bool = False,
    **kwargs,
) -> Path:
    """
    Generate a steering vector using SAE decoder features.

    Args:
        task: lm-eval task name (e.g., 'boolq', 'cb')
        model_name: HuggingFace model name (must be Gemma 2B or 9B)
        output_path: Where to save the steering vector
        trait_label: Label for the trait being steered
        num_pairs: Number of contrastive pairs to use
        method: Method name (should be 'sae')
        layers: Layer(s) to use (e.g., '12' or '10,11,12')
        normalize: Whether to normalize the steering vector
        device: Device to run on
        keep_intermediate: Whether to keep intermediate files
        top_k: Number of top SAE features to use

    Returns:
        Path to the saved steering vector
    """
    output_path = Path(output_path)

    if model_name not in SAE_CONFIGS:
        raise ValueError(
            f"No SAE config for model '{model_name}'. "
            f"Supported models: {list(SAE_CONFIGS.keys())}"
        )

    config = SAE_CONFIGS[model_name]

    # Parse layers
    if layers == "all":
        layer_indices = list(range(config["num_layers"]))
    else:
        layer_indices = [int(l.strip()) for l in layers.split(",")]

    # Step 1: Generate contrastive pairs
    print(f"Step 1: Generating contrastive pairs from task: {task}")
    pairs, pairs_file = generate_contrastive_pairs(task, num_pairs)
    print(f"   Loaded {len(pairs)} contrastive pairs")

    # Step 2: Load model
    print(f"\nStep 2: Loading model {model_name}...")
    model, tokenizer = load_model_and_tokenizer(model_name, device)

    steering_vectors = {}
    feature_info = {}

    for layer_idx in layer_indices:
        print(f"\nStep 3: Processing layer {layer_idx}")

        # Load SAE for this layer
        sae, sparsity = load_sae(model_name, layer_idx, device=device)

        # Step 4: Compute feature difference
        print(f"\nStep 4: Computing feature diff for layer {layer_idx}...")
        feat_diff = compute_feature_diff(model, tokenizer, sae, pairs, layer_idx, device)

        # Step 5: Compute steering vector from decoder
        print(f"\nStep 5: Computing steering vector from decoder for layer {layer_idx}...")
        steering_vec, feat_info = compute_steering_vector_from_decoder(
            feat_diff, sae, top_k=top_k, normalize=normalize
        )

        steering_vectors[str(layer_idx)] = steering_vec.cpu().float().tolist()
        feature_info[str(layer_idx)] = feat_info

        # Cleanup SAE
        del sae, sparsity, feat_diff
        gc.collect()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

    # Cleanup model
    del model
    gc.collect()
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        torch.cuda.synchronize()

    # Cleanup temp files
    if not keep_intermediate:
        import os
        os.unlink(pairs_file)

    # Save results
    result = {
        "steering_vectors": steering_vectors,
        "layers": [str(l) for l in layer_indices],
        "model": model_name,
        "method": "sae",
        "trait_label": trait_label,
        "task": task,
        "num_pairs": len(pairs),
        "sae_config": {
            "release": config["sae_release"],
            "top_k": top_k,
            "normalize": normalize,
        },
        "feature_info": feature_info,
    }

    output_path.parent.mkdir(parents=True, exist_ok=True)



"""
Pre-flight checks for steering method selection.

Analyzes activation geometry and reports raw geometry scores for the
chosen steering method. Does not make unvalidated method recommendations.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, Any
import torch

from wisent.core.primitives.contrastive_pairs.diagnostics.control_vectors import (
    detect_geometry_structure,
    GeometryAnalysisConfig,
    GeometryAnalysisResult,
    StructureType,
)
from wisent.core.utils.config_tools.constants import (
    SEPARATOR_WIDTH_STANDARD,
    SCORE_RANGE_MIN,
)
from wisent.core.control.steering_methods._helpers.preflight_helpers import (
    PreflightThresholds,
    complete_method_compatibility_check,
)

__all__ = [
    "PreflightCheckResult",
    "PreflightWarning",
    "run_preflight_check",
    "check_method_compatibility",
]

# Method-to-structure mapping: which geometry metric is relevant for each method.
# This is NOT a heuristic — it just selects which raw score to report.
_METHOD_STRUCTURE_KEY: Dict[str, str] = {
    "caa": "linear",
    "mean_diff": "linear",
    "tecza": "cone",
    "grom": "manifold",
    "tetno": "bimodal",
}


@dataclass
class PreflightWarning:
    """A single warning from pre-flight check."""
    severity: str  # "info", "warning", "error"
    message: str
    details: Optional[str] = None
    suggestion: Optional[str] = None


@dataclass
class PreflightCheckResult:
    """Results from pre-flight steering method check."""
    chosen_method: str
    geometry_result: GeometryAnalysisResult
    is_compatible: bool
    compatibility_score: float
    warnings: List[PreflightWarning] = field(default_factory=list)
    recommended_methods: List[str] = field(default_factory=list)

    def has_warnings(self) -> bool:
        """Check if there are any warnings."""
        return bool(self.warnings)

    def has_errors(self) -> bool:
        """Check if there are any error-level warnings."""
        return any(w.severity == "error" for w in self.warnings)

    def print_report(self) -> None:
        """Print a human-readable report."""
        from wisent.core.control.steering_methods._helpers.preflight_helpers import print_preflight_report
        print_preflight_report(self)


def run_preflight_check(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    chosen_method: str,
    config: Optional[GeometryAnalysisConfig] = None,
    *,
    thresholds: PreflightThresholds,
) -> PreflightCheckResult:
    """Run pre-flight check for a steering method."""
    geo_result = detect_geometry_structure(pos_activations, neg_activations, config)
    is_compatible, compat_score, warnings = check_method_compatibility(
        chosen_method, geo_result, thresholds=thresholds,
    )
    return PreflightCheckResult(
        chosen_method=chosen_method,
        geometry_result=geo_result,
        is_compatible=is_compatible,
        compatibility_score=compat_score,
        warnings=warnings,
        recommended_methods=[],
    )


def check_method_compatibility(
    method: str,
    geo_result: GeometryAnalysisResult,
    *,
    thresholds: PreflightThresholds,
) -> Tuple[bool, float, List[PreflightWarning]]:
    """Check if a steering method is compatible with detected geometry.

    Returns raw geometry scores instead of transforming them through
    unvalidated heuristic decision trees. The compatibility score is
    the raw geometry score for the structure relevant to the method.
    """
    warnings: List[PreflightWarning] = []
    method_lower = method.lower()
    structure_scores = {name: s.score for name, s in geo_result.all_scores.items()}

    structure_key = _METHOD_STRUCTURE_KEY.get(method_lower)

    if structure_key is not None:
        compat_score = structure_scores.get(structure_key, SCORE_RANGE_MIN)
        score_details = ", ".join(f"{k}={v:.2f}" for k, v in structure_scores.items())
        warnings.append(PreflightWarning(
            severity="info",
            message=f"Geometry scores for method '{method}'",
            details=f"Relevant score ({structure_key}): {compat_score:.2f}. "
                    f"All scores: {score_details}",
        ))
        is_compatible = compat_score >= thresholds.compat_min_compatible
        return is_compatible, compat_score, warnings

    # Unknown method — delegate to helper
    return complete_method_compatibility_check(
        method, geo_result.best_structure, structure_scores, warnings,
        preflight_compat_min_compatible=thresholds.compat_min_compatible,
    )



"""Advanced steering objects (TECZA, TETNO)."""
from __future__ import annotations
from typing import Any, Dict, List, Optional
import torch
from wisent.core.control.steering_methods._steering_object_base import (
    BaseSteeringObject, SteeringObjectMetadata, LayerName,
)
from wisent.core.control.steering_methods.configs.optimal import get_optimal

class TECZASteeringObject(BaseSteeringObject):
    """
    TECZA steering object with multiple directions per layer.
    
    Stores all discovered directions and their weights,
    allowing flexible combination strategies.
    """
    
    method_name = "tecza"
    
    def __init__(
        self,
        metadata: SteeringObjectMetadata,
        directions: Dict[int, torch.Tensor],  # layer -> [num_directions, hidden_dim]
        direction_weights: Optional[Dict[int, torch.Tensor]] = None,  # layer -> [num_directions]
        primary_only: bool = True,
    ):
        super().__init__(metadata)
        self.directions = directions
        self.direction_weights = direction_weights or {}
        self.primary_only = primary_only
    
    def get_steering_vector(self, layer: int, hidden_state: Optional[torch.Tensor] = None) -> torch.Tensor:
        if layer not in self.directions:
            raise KeyError(f"No directions for layer {layer}")
        
        dirs = self.directions[layer]
        
        if self.primary_only or layer not in self.direction_weights:
            return dirs[0]  # Primary direction
        
        # Weighted combination
        weights = self.direction_weights[layer]
        weighted = (weights.unsqueeze(-1) * dirs).sum(dim=0)
        return F.normalize(weighted, dim=-1)
    
    def compute_gate(self, hidden_state: torch.Tensor) -> torch.Tensor:
        batch_size = hidden_state.shape[0] if hidden_state.dim() > 1 else 1
        return torch.ones(batch_size, device=hidden_state.device, dtype=hidden_state.dtype)
    
    def compute_intensity(self, hidden_state: torch.Tensor, layer: int) -> torch.Tensor:
        batch_size = hidden_state.shape[0] if hidden_state.dim() > 1 else 1
        return torch.ones(batch_size, device=hidden_state.device, dtype=hidden_state.dtype)
    
    def get_all_directions(self, layer: int) -> torch.Tensor:
        """Get all directions for a layer."""
        return self.directions[layer]
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            'method': self.method_name,
            'metadata': {
                'method': self.metadata.method,
                'model_name': self.metadata.model_name,
                'benchmark': self.metadata.benchmark,
                'category': self.metadata.category,
                'extraction_strategy': self.metadata.extraction_strategy,
                'num_pairs': self.metadata.num_pairs,
                'layers': self.metadata.layers,
                'hidden_dim': self.metadata.hidden_dim,
                'created_at': self.metadata.created_at,
                'extra': self.metadata.extra,
                'extraction_component': self.metadata.extraction_component,
            },
            'directions': {str(k): v for k, v in self.directions.items()},
            'direction_weights': {str(k): v for k, v in self.direction_weights.items()},
            'primary_only': self.primary_only,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "TECZASteeringObject":
        meta_data = data['metadata']
        metadata = SteeringObjectMetadata(
            method=meta_data['method'],
            model_name=meta_data['model_name'],
            benchmark=meta_data['benchmark'],
            category=meta_data['category'],
            extraction_strategy=meta_data['extraction_strategy'],
            num_pairs=meta_data['num_pairs'],
            layers=meta_data['layers'],
            hidden_dim=meta_data['hidden_dim'],
            created_at=meta_data.get('created_at', ''),
            extra=meta_data.get('extra', {}),
            extraction_component=meta_data.get('extraction_component', get_optimal("extraction_component")),
        )

        def to_tensor(v):
            return torch.tensor(v) if isinstance(v, list) else v

        directions = {int(k): to_tensor(v) for k, v in data['directions'].items()}
        direction_weights = {int(k): to_tensor(v) for k, v in data.get('direction_weights', {}).items()}

        return cls(
            metadata=metadata,
            directions=directions,
            direction_weights=direction_weights,
            primary_only=data.get('primary_only', True),
        )


class TETNOSteeringObject(BaseSteeringObject):
    """
    TETNO steering object with conditional gating.
    
    Includes:
    - Behavior vectors per layer
    - Condition vector at sensor layer
    - Learned threshold for gating
    - Per-layer scaling factors
    """
    
    method_name = "tetno"
    
    def __init__(
        self,
        metadata: SteeringObjectMetadata,
        behavior_vectors: Dict[int, torch.Tensor],
        condition_vector: torch.Tensor,
        sensor_layer: int,
        threshold: float,
        layer_scales: Dict[int, float],
        gate_temperature: float,
    ):
        super().__init__(metadata)
        self.behavior_vectors = behavior_vectors
        self.condition_vector = condition_vector
        self.sensor_layer = sensor_layer
        self.threshold = threshold
        self.layer_scales = layer_scales
        self.gate_temperature = gate_temperature
    
    def get_steering_vector(self, layer: int, hidden_state: Optional[torch.Tensor] = None) -> torch.Tensor:
        if layer not in self.behavior_vectors:
            raise KeyError(f"No behavior vector for layer {layer}")
        return self.behavior_vectors[layer]
    
    def compute_gate(self, hidden_state: torch.Tensor) -> torch.Tensor:
        """Compute gate based on similarity to condition vector."""
        h_norm = F.normalize(hidden_state, dim=-1)
        c_norm = F.normalize(self.condition_vector.to(device=hidden_state.device, dtype=hidden_state.dtype), dim=-1)
        
        similarity = (h_norm * c_norm).sum(dim=-1)
        gate = torch.sigmoid((similarity - self.threshold) / self.gate_temperature)
        return gate
    
    def compute_intensity(self, hidden_state: torch.Tensor, layer: int) -> torch.Tensor:
        batch_size = hidden_state.shape[0] if hidden_state.dim() > 1 else 1
        if layer not in self.layer_scales:
            raise KeyError(f"No layer_scale for layer {layer}")
        scale = self.layer_scales[layer]
        return torch.full((batch_size,), scale, device=hidden_state.device, dtype=hidden_state.dtype)
    
    def should_steer(self, hidden_state: torch.Tensor) -> bool:
        """Hard decision on whether to steer."""
        gate = self.compute_gate(hidden_state)
        return gate.mean().item() > 0.5
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            'method': self.method_name,
            'metadata': {
                'method': self.metadata.method,
                'model_name': self.metadata.model_name,
                'benchmark': self.metadata.benchmark,
                'category': self.metadata.category,
                'extraction_strategy': self.metadata.extraction_strategy,
                'num_pairs': self.metadata.num_pairs,
                'layers': self.metadata.layers,
                'hidden_dim': self.metadata.hidden_dim,
                'created_at': self.metadata.created_at,
                'extra': self.metadata.extra,
                'extraction_component': self.metadata.extraction_component,
            },
            'behavior_vectors': {str(k): v for k, v in self.behavior_vectors.items()},
            'condition_vector': self.condition_vector,
            'sensor_layer': self.sensor_layer,
            'threshold': self.threshold,
            'layer_scales': {str(k): v for k, v in self.layer_scales.items()},
            'gate_temperature': self.gate_temperature,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "TETNOSteeringObject":
        meta_data = data['metadata']
        metadata = SteeringObjectMetadata(
            method=meta_data['method'],
            model_name=meta_data['model_name'],
            benchmark=meta_data['benchmark'],
            category=meta_data['category'],
            extraction_strategy=meta_data['extraction_strategy'],
            num_pairs=meta_data['num_pairs'],
            layers=meta_data['layers'],
            hidden_dim=meta_data['hidden_dim'],
            created_at=meta_data.get('created_at', ''),
            extra=meta_data.get('extra', {}),
            extraction_component=meta_data.get('extraction_component', get_optimal("extraction_component")),
        )

        def to_tensor(v):
            return torch.tensor(v) if isinstance(v, list) else v

        behavior_vectors = {int(k): to_tensor(v) for k, v in data['behavior_vectors'].items()}
        condition_vector = to_tensor(data['condition_vector'])
        layer_scales = {int(k): float(v) for k, v in data['layer_scales'].items()}
        
        return cls(
            metadata=metadata,
            behavior_vectors=behavior_vectors,
            condition_vector=condition_vector,
            sensor_layer=data['sensor_layer'],
            threshold=data['threshold'],
            layer_scales=layer_scales,
            gate_temperature=data['gate_temperature'],
        )





"""WICHER (Whirlwind) steering method constants."""
from wisent.core.utils.config_tools.constants.for_experiments.by_method.simple import *  # noqa: F401,F403

# --- WICHER code-enforced bounds ---
SS_WICHER_CONCEPT_DIMS = (0, 8, 16, 32, 64)
SS_WICHER_VAR_THRESH_LOW = 0.0
SS_WICHER_VAR_THRESH_HIGH = 1.0
SS_WICHER_NUM_STEPS_MIN = 1
SS_WICHER_NUM_STEPS_MAX = 20
SS_WICHER_ETA_LOW = -1.0
SS_WICHER_ETA_HIGH = 1.0
SS_WICHER_BETA_LOW = 0.0
SS_WICHER_BETA_HIGH = 1.0
SS_WICHER_ALPHA_DECAY_LOW = 0.0
SS_WICHER_ALPHA_DECAY_HIGH = 1.0

# --- WICHER concept subspace bounds ---
WICHER_CONCEPT_DIM_MIN = 1
WICHER_CONCEPT_DIM_MAX = 256

# --- WICHER solver choices ---
WICHER_SOLVER_BROYDEN = "broyden"
WICHER_SOLVER_NEWTON = "newton"
WICHER_SOLVER_HALLEY = "halley"
WICHER_SOLVER_CHOICES = (WICHER_SOLVER_BROYDEN, WICHER_SOLVER_NEWTON, WICHER_SOLVER_HALLEY)
WICHER_DEFAULT_SOLVER = WICHER_SOLVER_BROYDEN

# --- Halley correction factors ---
HALLEY_CORRECTION_HALF = 0.5
HALLEY_HESSIAN_DOT_FACTOR = 2
HALLEY_HESSIAN_NORM_FACTOR = 3




"""Phase 2a-2b: Benchmark and personalization steering optimization."""
import argparse
import json
import os

from wisent.core.control.steering_methods.configs.optimal import get_optimal
from wisent.core.utils.cli.optimization.core.optimize_helpers import save_checkpoint
from wisent.core.utils.config_tools.constants import (
    DISPLAY_TRUNCATION_ERROR,
)


def run_benchmark_steering(args, benchmarks, results):
    """Run Phase 2a: benchmark steering and 2b: personalization steering.
    
    Modifies results dict in-place.
    """
    from wisent.core.utils.cli.optimize_steering import execute_optimize_steering

    
    for bench_idx, benchmark in enumerate(benchmarks, 1):
        # Skip if already in checkpoint results
        if benchmark in results.get("steering", {}):
            print(f"\n   [{bench_idx}/{len(benchmarks)}] {benchmark} - SKIPPED (in checkpoint)")
            continue
        
        # Skip if already optimized (unless --force)
        if not getattr(args, 'force', False):
            cached = get_cached_optimization(args.model, benchmark, method="*")
            if cached:
                print(f"\n   [{bench_idx}/{len(benchmarks)}] {benchmark} - SKIPPED (cached)")
                results["steering"][benchmark] = {
                    "best_method": cached.method,
                    "best_layer": cached.layer,
                    "best_score": cached.score,
                    "from_cache": True,
                }
                continue
        
        print(f"\n   [{bench_idx}/{len(benchmarks)}] {benchmark}")
        
        try:
            steering_args = argparse.Namespace(
                model=args.model,
                steering_action="comprehensive",
                tasks=[benchmark],
                methods=methods,
                device=args.device,
                use_cached=False,
                save_as_default=True,
                compute_baseline=True,
            )
            
            steering_result = execute_optimize_steering(steering_args)
            
            if steering_result and benchmark in steering_result:
                best_result = steering_result[benchmark]
                # Get best method result
                best_method = None
                best_score = -1
                for method, method_result in best_result.items():
                    if isinstance(method_result, dict) and method_result["best_score"] > best_score:
                        best_score = method_result["best_score"]
                        best_method = method
                        best_layer = method_result.get("best_layer")
                        best_strength = method_result.get("best_strength")
                        if best_layer is None:
                            raise ValueError(f"Missing 'best_layer' in result for {method}")
                        if best_strength is None:
                            raise ValueError(f"Missing 'best_strength' in result for {method}")
                
                if best_method:
                    store_optimization(
                        model=args.model,
                        task=benchmark,
                        layer=best_layer,
                        strength=best_strength,
                        method=best_method.upper(),
                        strategy=get_optimal("steering_strategy"),
                        score=best_score,
                        metric="accuracy",
                    )
                    
                    results["steering"][benchmark] = {
                        "best_method": best_method,
                        "best_layer": best_layer,
                        "best_strength": best_strength,
                        "best_score": best_score,
                    }
                    print(f"       Best: {best_method} @ layer {best_layer} = {best_score:.3f}")
            
        except Exception as e:
            error_msg = f"{benchmark}: {str(e)}"
            results["errors"].append(error_msg)
            print(f"       Error: {str(e)[:DISPLAY_TRUNCATION_ERROR]}")
            logger.exception(f"Error optimizing {benchmark}")
        
        save_checkpoint(args.model, results, phase=f"steering_benchmark_{bench_idx}")
    
    # 2b. Personalization trait steering
    if personalization_traits:
        print(f"\n   Optimizing steering for {len(personalization_traits)} personalization traits...")
        
        for trait_idx, trait in enumerate(personalization_traits, 1):
            task_key = f"trait:{trait}"
            
            if task_key in results.get("steering", {}):
                print(f"\n   [{trait_idx}/{len(personalization_traits)}] {trait} - SKIPPED (in checkpoint)")
                continue
            
            if not getattr(args, 'force', False):
                cached = get_cached_optimization(args.model, f"personalization:{trait}", method="*")
                if cached:
                    print(f"\n   [{trait_idx}/{len(personalization_traits)}] {trait} - SKIPPED (cached)")
                    results["steering"][task_key] = {
                        "best_method": cached.method,
                        "best_layer": cached.layer,
                        "best_score": cached.score,
                        "from_cache": True,
                    }
                    continue
            
            print(f"\n   [{trait_idx}/{len(personalization_traits)}] {trait}")
            
            try:
                steering_args = argparse.Namespace(
                    model=args.model,
                    steering_action="personalization",
                    trait=trait,
                    methods=methods,
                    device=args.device,
                )
                
                steering_result = execute_optimize_steering(steering_args)
                
                if steering_result:
                    best_method = steering_result.get("best_method", "CAA")
                    best_layer = steering_result.get("best_layer")
                    if best_layer is None:
                        raise ValueError(f"Missing 'best_layer' in personalization result for {trait}")
                    best_score = steering_result["best_score"]
                    
                    store_optimization(
                        model=args.model,
                        task=f"personalization:{trait}",
                        layer=best_layer,
                        strength=steering_result["best_strength"],
                        method=best_method.upper(),
                        score=best_score,
                    )
                    
                    results["steering"][task_key] = {
                        "best_method": best_method,
                        "best_layer": best_layer,
                        "best_score": best_score,
                    }
                    print(f"       Best: {best_method} @ layer {best_layer} = {best_score:.3f}")
                
            except Exception as e:
                error_msg = f"trait:{trait}: {str(e)}"
                results["errors"].append(error_msg)
                print(f"       Error: {str(e)[:DISPLAY_TRUNCATION_ERROR]}")

            save_checkpoint(args.model, results, phase=f"steering_personalization_{trait_idx}")
    



"""Per-method optimization runner for comprehensive comparison."""
from __future__ import annotations

import json
import logging
import os
import tempfile
from typing import Any, Dict, List, Optional

from wisent.core.utils.config_tools.constants import (
    JSON_INDENT,
    SCORE_RANGE_MIN,
    TRIALS_PER_DIMENSION_MULTIPLIER,
)

logger = logging.getLogger(__name__)


def run_method_search(
    model: str, task_name: str, method: str,
    train_pairs_file: str, test_pairs_file: str,
    num_layers: int,
    device: Optional[str], verbose: bool,
    backend: str,
    search_overrides: Dict[str, Any],
    early_rejection_config: Dict[str, Any],
    output_dir: Optional[str] = None,
) -> Dict[str, Any]:
    """Run optimization for a single method.

    Uses the BaseOptimizer with the specified backend (hyperopt or
    optuna). The number of trials is derived from the search space
    dimensionality: len(space) * TRIALS_PER_DIMENSION_MULTIPLIER.

    Trains on train_pairs_file and evaluates on test_pairs_file to
    prevent data leakage.
    """
    from wisent.core.utils.cli.optimize_steering.search_space import (
        get_method_space,
    )
    from wisent.core.utils.cli.optimize_steering.pipeline import (
        create_objective,
    )
    from wisent.core.utils.services.optimization.core.atoms import (
        BaseOptimizer,
        HPOConfig,
    )

    space = get_method_space(method, num_layers)
    space = _apply_search_overrides(space, search_overrides)

    n_trials = len(space) * TRIALS_PER_DIMENSION_MULTIPLIER

    if verbose:
        print(f"   {method}: {n_trials} total configurations "
              f"(backend={backend})")

    optimizer = BaseOptimizer()
    optimizer.direction = "maximize"

    with tempfile.TemporaryDirectory() as work_dir:
        raw_objective = create_objective(
            method=method, model=model, task=task_name,
            num_layers=num_layers, limit=None, device=device,
            work_dir=work_dir, train_pairs_file=train_pairs_file,
            test_pairs_file=test_pairs_file,
        )
        if early_rejection_config["enabled"]:
            objective = _wrap_with_early_rejection(
                raw_objective, early_rejection_config,
            )
        else:
            objective = raw_objective
        result = optimizer.optimize_fn(objective, space, n_trials, cfg=HPOConfig(backend=backend))

    if verbose:
        print(f"   {method}: best={result.best_score:.4f} "
              f"({len(result.all_trials)} trials)")

    summary = {
        "method": method, "best_score": result.best_score,
        "best_params": result.best_params,
        "all_trials": result.all_trials,
        "n_trials": n_trials, "backend": backend,
    }

    if output_dir:
        _replay_best_config(
            model, task_name, method, result.best_params,
            train_pairs_file, test_pairs_file,
            device, verbose, output_dir, summary,
        )

    return summary


def _apply_search_overrides(
    space: Dict, overrides: Dict,
) -> Dict:
    """Apply user search overrides to the Optuna search space."""
    from wisent.core.utils.services.optimization.core.parameters import (
        CategoricalParam,
    )
    if "search_layers" in overrides and overrides["search_layers"]:
        layers = [
            int(x) for x in overrides["search_layers"].split(",")
        ]
        space["layer"] = CategoricalParam(choices=layers)
    if "search_strengths" in overrides and overrides["search_strengths"]:
        strengths = [
            float(x) for x in overrides["search_strengths"].split(",")
        ]
        space["strength"] = CategoricalParam(choices=strengths)
    if "search_strategies" in overrides and overrides["search_strategies"]:
        strats = [
            s.strip()
            for s in overrides["search_strategies"].split(",")
        ]
        space["steering_strategy"] = CategoricalParam(choices=strats)
    return space


def _replay_best_config(
    model: str, task_name: str, method: str,
    best_params: Dict[str, Any],
    train_pairs_file: str, test_pairs_file: str,
    device: Optional[str], verbose: bool,
    output_dir: str, summary: Dict[str, Any],
) -> None:
    """Re-run the best config and persist responses + scores."""
    from wisent.core.utils.cli.optimize_steering.pipeline import (
        run_pipeline, _build_config,
    )

    method_dir = os.path.join(output_dir, method.lower())
    os.makedirs(method_dir, exist_ok=True)

    if verbose:
        print(f"   Replaying best {method} config to persist responses...")

    config, strength = _build_config(method, best_params)
    result = run_pipeline(
        model=model, task=task_name, config=config,
        work_dir=method_dir, strength=strength,
        limit=None, device=device,
        train_pairs_file=train_pairs_file,
        test_pairs_file=test_pairs_file,
    )

    summary_path = os.path.join(method_dir, "optimization_summary.json")
    with open(summary_path, "w") as f:
        json.dump(summary, f, indent=JSON_INDENT, default=str)

    if verbose:
        print(f"   Saved responses to {method_dir}/responses.json")
        print(f"   Saved scores to {method_dir}/scores.json")


def _wrap_with_early_rejection(objective_fn, config: Dict[str, Any]):
    """Wrap objective to return SCORE_RANGE_MIN for rejected trials."""
    cv_threshold = config["cv_threshold"]

    def wrapped(params: dict) -> float:
        score = objective_fn(params)
        if score < cv_threshold:
            logger.info(
                f"   Early rejection: score {score:.4f} < "
                f"cv_threshold {cv_threshold}",
            )
            return SCORE_RANGE_MIN
        return score

    return wrapped



"""Method configuration dataclasses for steering optimization."""
from dataclasses import dataclass, field
from typing import List, Dict, Any, Optional
from typing import Optional as _Optional

from wisent.core.control.steering_optimizer.types import SteeringApplicationStrategy

STEERING_STRATEGIES = [s.value for s in SteeringApplicationStrategy]


@dataclass
class MethodConfig:
    """Base configuration for all methods."""
    method: str = ""
    extraction_strategy: Optional[str] = None
    steering_strategy: Optional[str] = None
    _extra_args: Dict[str, Any] = field(default_factory=dict)


@dataclass
class CAAConfig(MethodConfig):
    """CAA-specific parameters."""
    layer: _Optional[int] = None

    def to_args(self) -> Dict[str, Any]:
        return {"method": "CAA", "layer": self.layer,
                "steering_strategy": self.steering_strategy, **self._extra_args}


@dataclass
class OstrzeConfig(MethodConfig):
    """Ostrze-specific parameters."""
    layer: _Optional[int] = None

    def to_args(self) -> Dict[str, Any]:
        return {"method": "Ostrze", "layer": self.layer,
                "steering_strategy": self.steering_strategy, **self._extra_args}


@dataclass
class MLPConfig(MethodConfig):
    """MLP-specific parameters."""
    layer: _Optional[int] = None
    hidden_dim: _Optional[int] = None
    num_layers: _Optional[int] = None
    mlp_input_divisor: _Optional[int] = None
    mlp_early_stopping_patience: _Optional[int] = None
    mlp_gating_hidden_dim_divisor: _Optional[int] = None

    def to_args(self) -> Dict[str, Any]:
        base = {"method": "MLP", "layer": self.layer, "mlp_hidden_dim": self.hidden_dim,
                "mlp_num_layers": self.num_layers, "steering_strategy": self.steering_strategy,
                "mlp_input_divisor": self.mlp_input_divisor,
                "mlp_early_stopping_patience": self.mlp_early_stopping_patience,
                "mlp_gating_hidden_dim_divisor": self.mlp_gating_hidden_dim_divisor}
        base.update(self._extra_args)
        return base


@dataclass
class TECZAConfig(MethodConfig):
    """TECZA-specific parameters."""
    layer: _Optional[int] = None
    num_directions: _Optional[int] = None
    direction_weighting: _Optional[str] = None
    retain_weight: _Optional[float] = None
    optimization_steps: _Optional[int] = None

    def to_args(self) -> Dict[str, Any]:
        base = {"method": "TECZA", "layer": self.layer,
                "direction_weighting": self.direction_weighting,
                "steering_strategy": self.steering_strategy}
        base.update(self._extra_args)
        return base


@dataclass
class TETNOConfig(MethodConfig):
    """TETNO-specific parameters."""
    sensor_layer: _Optional[int] = None
    steering_layers: List[int] = field(default_factory=list)
    condition_threshold: _Optional[float] = None
    gate_temperature: _Optional[float] = None
    max_alpha: _Optional[float] = None

    def to_args(self) -> Dict[str, Any]:
        base = {"method": "TETNO", "sensor_layer": self.sensor_layer,
                "steering_layers": ",".join(str(l) for l in self.steering_layers),
                "steering_strategy": self.steering_strategy}
        base.update(self._extra_args)
        return base


@dataclass
class GROMConfig(MethodConfig):
    """GROM-specific parameters."""
    sensor_layer: _Optional[int] = None
    steering_layers: List[int] = field(default_factory=list)
    num_directions: _Optional[int] = None
    gate_hidden_dim: _Optional[int] = None
    intensity_hidden_dim: _Optional[int] = None
    behavior_weight: _Optional[float] = None
    retain_weight: _Optional[float] = None
    sparse_weight: _Optional[float] = None
    max_alpha: _Optional[float] = None
    optimization_steps: _Optional[int] = None

    def to_args(self) -> Dict[str, Any]:
        base = {"method": "GROM", "sensor_layer": self.sensor_layer,
                "steering_layers": ",".join(str(l) for l in self.steering_layers),
                "steering_strategy": self.steering_strategy}
        base.update(self._extra_args)
        return base


@dataclass
class NurtConfig(MethodConfig):
    """Nurt-specific parameters."""
    layer: _Optional[int] = None
    num_dims: _Optional[int] = None
    variance_threshold: _Optional[float] = None
    training_epochs: _Optional[int] = None
    lr: _Optional[float] = None
    num_integration_steps: _Optional[int] = None
    t_max: _Optional[float] = None

    def to_args(self) -> Dict[str, Any]:
        base = {"method": "nurt", "layer": self.layer,
                "steering_strategy": self.steering_strategy}
        base.update(self._extra_args)
        return base


@dataclass
class SzlakConfig(MethodConfig):
    """Szlak-specific parameters."""
    layer: _Optional[int] = None
    sinkhorn_reg: _Optional[float] = None
    inference_k: _Optional[int] = None

    def to_args(self) -> Dict[str, Any]:
        base = {"method": "szlak", "layer": self.layer,
                "steering_strategy": self.steering_strategy}
        base.update(self._extra_args)
        return base


@dataclass
class WicherConfig(MethodConfig):
    """WICHER-specific parameters."""
    layer: _Optional[int] = None
    concept_dim: _Optional[int] = None
    variance_threshold: _Optional[float] = None
    num_steps: _Optional[int] = None
    alpha: _Optional[float] = None
    eta: _Optional[float] = None
    beta: _Optional[float] = None
    alpha_decay: _Optional[float] = None

    def to_args(self) -> Dict[str, Any]:
        base = {"method": "wicher", "layer": self.layer,
                "steering_strategy": self.steering_strategy}
        base.update(self._extra_args)
        return base





"""Mmmu group task manifest."""

from __future__ import annotations

BASE_IMPORT = "wisent.extractors.lm_eval.lm_task_extractors."

MMMU_TASKS = {
    "mmmu": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_accounting": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_agriculture": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_architecture_and_engineering": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_art": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_art_and_design": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_art_theory": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_basic_medical_science": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_biology": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_business": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_chemistry": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_clinical_medicine": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_computer_science": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_design": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_diagnostics_and_laboratory_medicine": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_economics": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_electronics": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_energy_and_power": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_finance": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_geography": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_health_and_medicine": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_history": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_humanities_and_social_science": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_literature": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_manage": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_marketing": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_materials": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_math": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_mechanical_engineering": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_music": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_pharmacy": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_physics": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_psychology": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_public_health": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_science": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_sociology": f"{BASE_IMPORT}mmmu:MmmuExtractor",
    "mmmu_val_tech_and_engineering": f"{BASE_IMPORT}mmmu:MmmuExtractor",
}



"""Flores group task manifest."""

from __future__ import annotations

from wisent.extractors.lm_eval.group_task_manifests.flores_languages_a_to_e import FLORES_TASKS_A_TO_E
from wisent.extractors.lm_eval.group_task_manifests.flores_languages_f_to_z import FLORES_TASKS_F_TO_Z

FLORES_TASKS = {**FLORES_TASKS_A_TO_E, **FLORES_TASKS_F_TO_Z}



"""Zwiad with automatic concept decomposition and naming."""
from typing import Dict, Any, Optional, List, Tuple, Set
from pathlib import Path
from dataclasses import asdict as _asdict
import json
import torch
import numpy as np
from wisent.core.utils.config_tools.constants import (
    DEFAULT_RANDOM_SEED, DIAGNOSTICS_TOTAL_CHECKS, JSON_INDENT,
    VIZ_PCA_COMPONENTS,
)
from wisent.core.reading.modules.utilities.metrics.core.metrics_core import compute_geometry_metrics
from wisent.core.reading.modules.utilities.concepts import decompose_and_name_concepts_with_labels, find_optimal_layer_per_concept
from wisent.core.reading.modules.utilities.data.database_loaders import load_activations_from_database, load_pair_texts_from_database, load_available_layers_from_database

_GEO_KEYS = ["linear_accuracy", "nonlinear_accuracy", "gap", "diagnosis", "confidence",
    "p_value", "gap_ci_lower", "gap_ci_upper", "n_diagnostics_passed", "n_diagnostics_total",
    "t_statistic", "residual_silhouette", "residuals_cluster", "ramsey_improvement",
    "ramsey_significant", "diagnostics"]


def _save_checkpoint(results: Dict, output_path: Optional[str]) -> None:
    if not output_path:
        return
    def _ser(obj):
        if isinstance(obj, torch.Tensor): return obj.tolist()
        if isinstance(obj, np.ndarray): return obj.tolist()
        if isinstance(obj, dict): return {str(k): _ser(v) for k, v in obj.items()}
        if isinstance(obj, (list, tuple)): return [_ser(v) for v in obj]
        if hasattr(obj, 'item'): return obj.item()
        return obj
    p = Path(output_path)
    p.parent.mkdir(parents=True, exist_ok=True)
    with open(p, 'w') as f:
        json.dump(_ser(results), f, indent=JSON_INDENT)


def _load_checkpoint(output_path: Optional[str]) -> Dict:
    if not output_path:
        return {}
    p = Path(output_path)
    if not p.exists() or p.stat().st_size == 0:
        return {}
    with open(p) as f:
        return json.load(f)

__all__ = [
    "run_zwiad_with_concept_naming", "extract_pair_texts_from_enriched_pairs",
    "load_activations_from_database", "load_pair_texts_from_database",
    "load_available_layers_from_database", "find_optimal_layer_per_concept",
]


def run_zwiad_with_concept_naming(
    activations_by_layer: Dict[int, Tuple[torch.Tensor, torch.Tensor]],
    llm_model: str,
    concept_naming_n_samples: int,
    min_clusters: int,
    pair_texts: Optional[Dict[int, Dict[str, str]]] = None,
    generate_visualizations: bool = True,
    steps: Optional[str] = None, output_path: Optional[str] = None,
    linearity_max_pairs: int = None, min_concept_dim: int = None,
    *,
    cv_folds: int,
    min_concept_pairs: int,
    zwiad_score_primary: float, zwiad_score_secondary: float, zwiad_score_tertiary: float,
    zwiad_editability_threshold: float, zwiad_przelom_bonus_max: float,
    probe_small_hidden: int, probe_mlp_hidden: int, probe_mlp_alpha: float,
    spectral_n_neighbors: int, direction_n_bootstrap: int, direction_subset_fraction: float,
    direction_std_penalty: float, consistency_w_cosine: float, consistency_w_positive: float,
    consistency_w_high_sim: float, sparsity_threshold_fraction: float,
    detection_threshold: float, direction_moderate_similarity: float,
) -> Dict[str, Any]:
    """Run Zwiad with geometry metrics, multi-step protocol, and concept naming."""
    if linearity_max_pairs is None:
        raise ValueError("linearity_max_pairs is required")
    if min_concept_dim is None:
        raise ValueError("min_concept_dim is required")
    from .zwiad_protocol import (
        test_signal, test_geometry, test_decomposition, select_intervention,
        test_editability, SignalTestResult, GeometryTestResult,
        DecompositionTestResult, EditabilityTestResult,
    )
    import time as _time
    import sys as _sys
    def _log(msg): print(f"  [TRACE] {msg}", file=_sys.stderr, flush=True)
    existing = _load_checkpoint(output_path)
    _log(f"Checkpoint loaded: {list(existing.keys())}")
    if steps is None:
        steps_to_run = {"signal", "geometry", "decomposition", "intervention", "editability"}
    else:
        steps_to_run = {s.strip().lower() for s in steps.split(",")}
    sorted_layers = sorted(activations_by_layer.keys())
    n_layers = len(sorted_layers)
    min_pairs = min(len(activations_by_layer[l][0]) for l in sorted_layers)
    pos_list = [activations_by_layer[l][0][:min_pairs] for l in sorted_layers]
    neg_list = [activations_by_layer[l][1][:min_pairs] for l in sorted_layers]
    pos_concat = torch.cat(pos_list, dim=1)
    neg_concat = torch.cat(neg_list, dim=1)
    n_pairs = len(pos_concat)
    from sklearn.decomposition import PCA as _PCA
    _pca_dims = min(n_pairs - 1, pos_concat.shape[1], VIZ_PCA_COMPONENTS)
    if _pca_dims < pos_concat.shape[1] and _pca_dims >= 2:
        _comb = torch.cat([pos_concat, neg_concat], dim=0).cpu().numpy()
        _comb_pca = _PCA(n_components=_pca_dims, random_state=DEFAULT_RANDOM_SEED).fit_transform(_comb)
        pos_pca = torch.tensor(_comb_pca[:n_pairs], dtype=pos_concat.dtype)
        neg_pca = torch.tensor(_comb_pca[n_pairs:], dtype=pos_concat.dtype)
    else:
        pos_pca, neg_pca = pos_concat, neg_concat
    if "metrics" in existing and existing["metrics"]:
        metrics = existing["metrics"]
    else:
        metrics = compute_geometry_metrics(pos_concat, neg_concat, min_clusters=min_clusters, n_folds=cv_folds, generate_visualizations=generate_visualizations, probe_small_hidden=probe_small_hidden, probe_mlp_hidden=probe_mlp_hidden, probe_mlp_alpha=probe_mlp_alpha, spectral_n_neighbors=spectral_n_neighbors, direction_n_bootstrap=direction_n_bootstrap, direction_subset_fraction=direction_subset_fraction, direction_std_penalty=direction_std_penalty, consistency_w_cosine=consistency_w_cosine, consistency_w_positive=consistency_w_positive, consistency_w_high_sim=consistency_w_high_sim, sparsity_threshold_fraction=sparsity_threshold_fraction, detection_threshold=detection_threshold, direction_moderate_similarity=direction_moderate_similarity)
    results = {
        "n_pairs": n_pairs, "n_layers": n_layers, "layers_used": sorted_layers,
        "total_dims": pos_concat.shape[1], "steps_run": list(steps_to_run),
        "metrics": metrics,
    }
    _save_checkpoint(results, output_path)
    signal_result, geometry_result, decomposition_result, editability_result = None, None, None, None
    min_pairs_for_probes = min_concept_dim
    # Step 1: Signal
    if "signal" in steps_to_run:
        if "signal_test" in existing:
            results["signal_test"] = existing["signal_test"]
            st = existing["signal_test"]
            if "max_z_score" in st:
                signal_result = SignalTestResult(
                    max_z_score=st["max_z_score"], min_p_value=st["min_p_value"],
                    passed=st["passed"], permutation_metrics=st.get("permutation_metrics", {}),
                    nonsense_metrics=st.get("nonsense_metrics"))
        elif n_pairs < min_pairs_for_probes:
            results["signal_test"] = {"status": "insufficient_data", "n_pairs": n_pairs}
        else:
            signal_result = test_signal(pos_pca, neg_pca, ["knn_accuracy", "knn_pca_accuracy", "mlp_probe_accuracy"], device=str(pos_pca.device))
            results["signal_test"] = {
                "max_z_score": signal_result.max_z_score, "min_p_value": signal_result.min_p_value,
                "passed": signal_result.passed, "permutation_metrics": signal_result.permutation_metrics,
                "nonsense_metrics": signal_result.nonsense_metrics}
        _save_checkpoint(results, output_path)
    # Step 2: Geometry
    if "geometry" in steps_to_run:
        if "geometry_test" in existing:
            results["geometry_test"] = existing["geometry_test"]
            gt = existing["geometry_test"]
            if "linear_accuracy" in gt:
                geometry_result = GeometryTestResult(**{k: gt[k] for k in _GEO_KEYS})
        elif n_pairs < min_pairs_for_probes:
            results["geometry_test"] = {"status": "insufficient_data", "n_pairs": n_pairs}
        else:
            _n_geo = min(n_pairs, linearity_max_pairs)
            if n_pairs > linearity_max_pairs:
                _idx = np.random.RandomState(DEFAULT_RANDOM_SEED).choice(n_pairs, linearity_max_pairs, replace=False)
                _idx.sort()
                _pg, _ng = pos_pca[_idx], neg_pca[_idx]
            else:
                _pg, _ng = pos_pca, neg_pca
            geometry_result = test_geometry(_pg, _ng, diagnostics_total_checks=DIAGNOSTICS_TOTAL_CHECKS)
            results["geometry_test"] = {k: getattr(geometry_result, k) for k in _GEO_KEYS}
        _save_checkpoint(results, output_path)
    # Step 3: Decomposition + Concept Naming
    if "decomposition" in steps_to_run:
        if "decomposition_test" in existing:
            results["decomposition_test"] = existing["decomposition_test"]
            dt = existing["decomposition_test"]
            if "n_concepts" in dt and "status" not in dt:
                decomposition_result = DecompositionTestResult(
                    n_concepts=dt["n_concepts"], cluster_labels=dt.get("cluster_labels", [0]*n_pairs),
                    silhouette_score=dt["silhouette_score"],
                    is_fragmented=dt.get("is_fragmented", False),
                    per_concept_sizes=dt.get("per_concept_sizes", {}))
            if "concept_decomposition" in existing:
                results["concept_decomposition"] = existing["concept_decomposition"]
        elif n_pairs < 3:
            results["decomposition_test"] = {"status": "insufficient_data", "n_pairs": n_pairs}
        else:
            decomposition_result = test_decomposition(pos_concat, neg_concat)
            results["decomposition_test"] = {
                "n_concepts": decomposition_result.n_concepts, "silhouette_score": decomposition_result.silhouette_score,
                "is_fragmented": decomposition_result.is_fragmented, "per_concept_sizes": decomposition_result.per_concept_sizes}
            decomp = decompose_and_name_concepts_with_labels(
                pos_concat, neg_concat, concept_naming_n_samples=concept_naming_n_samples,
                pair_texts=pair_texts, cluster_labels=decomposition_result.cluster_labels,
                n_concepts=decomposition_result.n_concepts,
                generate_visualizations=generate_visualizations, llm_model=llm_model)
            decomp["n_layers_used"] = n_layers
            labels_arr = np.array(decomposition_result.cluster_labels)
            pcl = find_optimal_layer_per_concept(activations_by_layer, labels_arr, decomposition_result.n_concepts, cv_folds=cv_folds, min_concept_pairs=min_concept_pairs)
            for concept in decomp.get("concepts", []):
                idx = concept["id"] - 1
                if idx in pcl:
                    concept["optimal_layer"] = pcl[idx]["best_layer"]
                    concept["optimal_layer_accuracy"] = pcl[idx]["best_accuracy"]
            decomp["per_concept_layers"] = pcl
            results["concept_decomposition"] = decomp
        _save_checkpoint(results, output_path)
    # Step 5: Editability
    if "editability" in steps_to_run:
        if "editability_test" in existing and existing["editability_test"]:
            results["editability_test"] = existing["editability_test"]
            et = existing["editability_test"]
            if "composite_editability" in et:
                editability_result = EditabilityTestResult(
                    composite_editability=et["composite_editability"],
                    steering_survival=et["steering_survival"],
                    spectral_concentration=et["spectral_concentration"],
                    spectral_sharpness=et["spectral_sharpness"],
                    attention_entropy=et["attention_entropy"],
                    jacobian_sensitivity=et["jacobian_sensitivity"])
        elif n_pairs >= 2:
            editability_result = test_editability(pos_pca, neg_pca)
            results["editability_test"] = {
                "composite_editability": editability_result.composite_editability,
                "steering_survival": editability_result.steering_survival,
                "spectral_concentration": editability_result.spectral_concentration,
                "spectral_sharpness": editability_result.spectral_sharpness,
                "attention_entropy": editability_result.attention_entropy,
                "jacobian_sensitivity": editability_result.jacobian_sensitivity}
        _save_checkpoint(results, output_path)
    # Step 4: Intervention Selection
    if "intervention" in steps_to_run and signal_result and geometry_result and decomposition_result:
        intervention = select_intervention(signal_result, geometry_result, decomposition_result, editability=editability_result,
            zwiad_score_primary=zwiad_score_primary, zwiad_score_secondary=zwiad_score_secondary,
            zwiad_score_tertiary=zwiad_score_tertiary, zwiad_editability_threshold=zwiad_editability_threshold,
            zwiad_przelom_bonus_max=zwiad_przelom_bonus_max)
        results["intervention"] = {
            "recommended_method": intervention.recommended_method, "confidence": intervention.confidence,
            "reasoning": intervention.reasoning, "method_scores": intervention.method_scores}
        results["recommended_method"] = intervention.recommended_method
        results["recommendation_confidence"] = intervention.confidence
        _save_checkpoint(results, output_path)
    return results


def extract_pair_texts_from_enriched_pairs(enriched_pairs: List[Dict]) -> Dict[int, Dict[str, str]]:
    """Extract pair texts from enriched pairs format (wisent CLI output)."""
    pair_texts = {}
    for i, pair in enumerate(enriched_pairs):
        pair_texts[i] = {
            "prompt": pair.get("prompt", ""),
            "positive": pair.get("positive_response", {}).get("model_response", ""),
            "negative": pair.get("negative_response", {}).get("model_response", ""),
        }
    return pair_texts



"""Helpers for intervention_validation_cli."""

import json
import re
import subprocess
import sys
from pathlib import Path
from dataclasses import dataclass
from typing import List, Dict, Any, Optional

from wisent.core.utils.config_tools.constants import (
    DISPLAY_TRUNCATION_LARGE,
)
from wisent.core.utils.infra_tools.infra.core.hardware import subprocess_timeout_long_s
from wisent.core.utils.infra_tools.errors import MissingParameterError

GCS_BUCKET = "wisent-images-bucket"
GCS_PREFIX = "intervention_validation"


class BenchmarkResult:
    benchmark: str
    diagnosis: str  # LINEAR, NONLINEAR, NO_SIGNAL
    layer: int
    
    # Baseline (no steering)
    baseline_accuracy: float
    
    # With steering
    steered_accuracy: float
    steering_strength: float
    
    # Effect
    accuracy_change: float
    steering_success: bool


def run_wisent_task(
    benchmark: str,
    model: str,
    layer: int,
    steering_strength: Optional[float] = None,
    steering_mode: bool = False,
) -> float:
    """
    Run wisent tasks command and return accuracy.
    """
    cmd = [
        sys.executable, "-m", "wisent.core.parser",
        "tasks", benchmark,
        "--model", model,
        "--layer", str(layer),
        "--output-mode", "likelihoods",
    ]
    
    if steering_mode:
        if steering_strength is None:
            raise MissingParameterError(
                params=["steering_strength"],
                context="run_wisent_task with steering_mode=True",
            )
        cmd.extend([
            "--steering-mode",
            "--steering-strength", str(steering_strength),
            "--steering-method", "CAA",
        ])
    
    print(f"  Running: {' '.join(cmd[-10:])}")
    
    try:
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            timeout=subprocess_timeout_long_s(),
        )
        
        output = result.stdout + result.stderr
        
        # Parse accuracy from output
        for line in output.split('\n'):
            if 'Test Accuracy:' in line or 'test_accuracy' in line.lower():
                # Try to extract number
                import re
                match = re.search(r'(\d+\.?\d*)\s*%?', line)
                if match:
                    acc = float(match.group(1))
                    if acc > 1:  # Percentage
                        acc /= 100
                    return acc
            if 'accuracy' in line.lower() and ':' in line:
                parts = line.split(':')
                if len(parts) >= 2:
                    try:
                        acc = float(parts[-1].strip().replace('%', ''))
                        if acc > 1:
                            acc /= 100
                        return acc
                    except:
                        pass
        
        print(f"  Warning: Could not parse accuracy from output")
        print(f"  Output: {output[:DISPLAY_TRUNCATION_LARGE]}")
        return 0.5
        
    except subprocess.TimeoutExpired:
        print(f"  Timeout running {benchmark}")
        return 0.5
    except Exception as e:
        print(f"  Error: {e}")
        return 0.5


def load_diagnosis_from_gcs(model_name: str) -> Dict[str, Any]:
    """Load Zwiad diagnosis results from GCS."""
    model_prefix = model_name.replace('/', '_')
    local_dir = Path(f"/tmp/diagnosis_{model_prefix}")
    local_dir.mkdir(parents=True, exist_ok=True)

    try:
        subprocess.run(
            ["gcloud", "storage", "rsync",
             f"gs://{GCS_BUCKET}/direction_discovery/{model_prefix}/",
             str(local_dir),
             "--quiet"],
            check=False,
            capture_output=True,
        )
    except Exception:
        pass
    
    results = {}
    for f in local_dir.glob(f"{model_prefix}_*.json"):
        if "summary" not in f.name:
            category = f.stem.replace(f"{model_prefix}_", "")
            with open(f) as fp:
                results[category] = json.load(fp)
    
    return results


def get_benchmarks_by_diagnosis(diagnosis_results: Dict) -> Dict[str, List[tuple]]:
    """Group benchmarks by diagnosis type. Returns {diagnosis: [(benchmark, layer, signal, linear)]}"""
    by_diagnosis = {"LINEAR": [], "NO_SIGNAL": [], "NONLINEAR": []}
    
    for category, data in diagnosis_results.items():
        results = data.get("results", [])
        seen = set()
        
        for r in results:
            bench = r["benchmark"]
            if bench in seen:
                continue
            seen.add(bench)
            
            signal = r["signal_strength"]
            linear = r["linear_probe_accuracy"]
            num_layers = len(r["layers"]) if r["layers"] else 36
            best_layer = num_layers // 2
            
            if signal < 0.6:
                by_diagnosis["NO_SIGNAL"].append((bench, best_layer, signal, linear))
            elif linear > 0.6 and (signal - linear) < 0.15:
                by_diagnosis["LINEAR"].append((bench, best_layer, signal, linear))
            else:
                by_diagnosis["NONLINEAR"].append((bench, best_layer, signal, linear))
    
    return by_diagnosis




"""Automatic parameter tuning for steering - no hardcoded defaults."""

import json
import numpy as np
import torch
from pathlib import Path
from typing import Dict, List, Tuple, Any
from datetime import datetime
from wisent.core.utils.infra_tools.errors import MissingParameterError
from wisent.core.utils.config_tools.constants import COMBO_OFFSET, RECURSION_INITIAL_DEPTH
from wisent.core.primitives.models.config import get_generate_kwargs

from wisent.core.utils.cli.steering_checkpoint import (
    SteeringConfig, AutotuneCheckpoint, log_config_result, setup_checkpoint_paths
)


def evaluate_steering_config(
    adapter,
    test_ids: List,
    pair_texts: Dict,
    evaluator,
    steering_vectors: Dict[str, torch.Tensor],
    layer_strengths: Dict[int, float],
    steering_method_name: str,
    pos_by_layer: Dict,
    neg_by_layer: Dict,
    max_new_tokens: int | None = None,
) -> int:
    """Evaluate a steering config on test set, return number of truthful responses."""
    from wisent.core.utils.cli.steering_behavioral import extract_response
    from wisent.core.utils.cli.steering_viz_helpers import create_steering_method
    from wisent.core.primitives.model_interface.core.activations.core.atoms import LayerActivations
    from wisent.core.primitives.model_interface.adapters.base import SteeringConfig as AdapterSteeringConfig

    scales_dict = {f"layer.{layer}": strength for layer, strength in layer_strengths.items() if strength > 0}
    methods_dict = {}
    for layer, strength in layer_strengths.items():
        if strength > 0:
            layer_name = f"layer.{layer}"
            pos_acts = pos_by_layer.get(layer)
            neg_acts = neg_by_layer.get(layer)
            if pos_acts is not None and neg_acts is not None:
                method = create_steering_method(steering_method_name, strength, pos_acts, neg_acts)
                if method is not None:
                    methods_dict[layer_name] = method

    active_vectors = {k: v for k, v in steering_vectors.items() if k in scales_dict}
    if not active_vectors:
        return 0

    steering_vecs = LayerActivations(active_vectors)
    config = AdapterSteeringConfig(scale=scales_dict, method=methods_dict if methods_dict else None)

    truthful_count = 0
    for pair_key in test_ids:
        pair_data = pair_texts[pair_key]
        prompt = pair_data.get("prompt", "")
        pos_ref_text = pair_data.get("positive", "")
        neg_ref_text = pair_data.get("negative", "")
        correct_answers = [pos_ref_text] if pos_ref_text else []
        incorrect_answers = [neg_ref_text] if neg_ref_text else []

        messages = [{"role": "user", "content": prompt}]
        formatted_prompt = adapter.apply_chat_template(messages, add_generation_prompt=True)
        steered_response = extract_response(
            adapter.forward_with_steering(formatted_prompt, steering_vectors=steering_vecs, config=config)
        )
        result = evaluator.evaluate(steered_response, pos_ref_text,
            correct_answers=correct_answers, incorrect_answers=incorrect_answers)
        if result.ground_truth == "TRUTHFUL":
            truthful_count += 1

    return truthful_count


def search_optimal_steering_config(
    adapter, train_ids: List, val_ids: List, pair_texts: Dict, evaluator,
    pos_by_layer: Dict, neg_by_layer: Dict, layer_accuracies: Dict[int, float],
    steering_vectors: Dict[str, torch.Tensor], max_new_tokens: int | None = None,
    checkpoint_dir: str = "./autotune_checkpoints", log_file: str = None,
    autotuner_strengths: tuple = None,
    autotuner_progress_interval: int = None,
    autotuner_accuracy_decrement: float = None,
    autotuner_accuracy_filter: float = None,
    autotuner_base_threshold: float = None,
) -> SteeringConfig:
    """Search for optimal steering parameters with checkpointing and resume capability."""
    for _pn, _pv in [("autotuner_strengths", autotuner_strengths), ("autotuner_progress_interval", autotuner_progress_interval),
                      ("autotuner_accuracy_decrement", autotuner_accuracy_decrement), ("autotuner_accuracy_filter", autotuner_accuracy_filter),
                      ("autotuner_base_threshold", autotuner_base_threshold)]:
        if _pv is None:
            raise ValueError(f"{_pn} is required")
    checkpoint_file, log_file_path = setup_checkpoint_paths(checkpoint_dir, log_file)

    acc_values = sorted(layer_accuracies.values())
    acc_thresholds = [autotuner_base_threshold] + [acc - autotuner_accuracy_decrement for acc in acc_values if acc > autotuner_accuracy_filter]
    strength_candidates = list(autotuner_strengths)
    method_candidates = ["linear", "clamping", "contrast"]
    total_combos = len(acc_thresholds) * len(strength_candidates) * len(method_candidates)

    all_configs = []
    for acc_threshold in acc_thresholds:
        for base_strength in strength_candidates:
            for method in method_candidates:
                all_configs.append((acc_threshold, base_strength, method))

    checkpoint = AutotuneCheckpoint.load(checkpoint_file)
    if checkpoint is not None:
        print(f"  Resuming from checkpoint: {checkpoint.last_combo_idx}/{checkpoint.total_combos} completed")
        best_score = checkpoint.best_score
        best_config = SteeringConfig.from_dict(checkpoint.best_config) if checkpoint.best_config else None
        start_idx = checkpoint.last_combo_idx
        completed_configs = checkpoint.completed_configs
    else:
        print(f"  Searching {total_combos} configurations (checkpoint: {checkpoint_file})")
        best_score = -1
        best_config = None
        start_idx = 0
        completed_configs = []

    for combo_idx in range(start_idx, len(all_configs)):
        acc_threshold, base_strength, method = all_configs[combo_idx]

        layer_strengths = {}
        for layer, acc in layer_accuracies.items():
            if acc > acc_threshold:
                max_acc = max(layer_accuracies.values())
                scale = (acc - acc_threshold) / (max_acc - acc_threshold) if max_acc > acc_threshold else 1.0
                layer_strengths[layer] = base_strength * scale
            else:
                layer_strengths[layer] = 0.0

        if not any(s > 0 for s in layer_strengths.values()):
            score = 0
        else:
            score = evaluate_steering_config(
                adapter, val_ids, pair_texts, evaluator, steering_vectors,
                layer_strengths, method, pos_by_layer, neg_by_layer, max_new_tokens
            )

        result = {
            "combo_idx": combo_idx + 1, "total": total_combos, "acc_threshold": acc_threshold,
            "base_strength": base_strength, "method": method, "score": score,
            "val_size": len(val_ids), "active_layers": sum(1 for s in layer_strengths.values() if s > 0),
            "timestamp": datetime.now().isoformat(),
        }
        completed_configs.append(result)
        log_config_result(log_file_path, result)

        if score > best_score:
            best_score = score
            best_config = SteeringConfig(
                layer_strengths=layer_strengths.copy(), steering_method=method,
                accuracy_threshold=acc_threshold, base_strength=base_strength,
                val_score=score, layer_accuracies=layer_accuracies.copy(),
            )
            active_layers = sum(1 for s in layer_strengths.values() if s > 0)
            print(f"    [{combo_idx+1}/{total_combos}] New best: {score}/{len(val_ids)} "
                  f"(thresh={acc_threshold:.2f}, str={base_strength}, method={method}, layers={active_layers})")
        elif (combo_idx + COMBO_OFFSET) % autotuner_progress_interval == RECURSION_INITIAL_DEPTH:
            print(f"    [{combo_idx+1}/{total_combos}] Progress... best so far: {best_score}/{len(val_ids)}")

        ckpt = AutotuneCheckpoint(
            completed_configs=completed_configs,
            best_config=best_config.to_dict() if best_config else None,
            best_score=best_score, last_combo_idx=combo_idx + 1,
            total_combos=total_combos, timestamp=datetime.now().isoformat(),
        )
        ckpt.save(checkpoint_file)

    print(f"  Search complete. Log saved to: {log_file_path}")
    print(f"  Final checkpoint: {checkpoint_file}")
    return best_config


def run_autotune_multilayer(
    adapter, train_ids, test_ids, all_pair_texts, pos_by_layer, neg_by_layer,
    behavioral_acts_by_layer, behavioral_labels, val_split, max_new_tokens
):
    """Run full autotune flow for multi-layer steering."""
    from wisent.core.utils.cli.steering_viz_helpers import (
        select_steering_direction, create_steering_method, create_all_layer_steering
    )
    from wisent.core.reading.evaluators.rotator import EvaluatorRotator

    EvaluatorRotator.discover_evaluators("wisent.core.reading.evaluators.oracles")
    EvaluatorRotator.discover_evaluators("wisent.core.reading.evaluators.benchmark_specific")

    available_layers = sorted(pos_by_layer.keys())

    steering_vectors_dict, _, _ = create_all_layer_steering(
        pos_by_layer, neg_by_layer, "linear", 1.0, "behavioral",
        behavioral_acts_by_layer=behavioral_acts_by_layer, behavioral_labels=behavioral_labels
    )

    layer_accuracies = {}
    for layer in available_layers:
        beh_acts = behavioral_acts_by_layer.get(layer)
        _, _, acc = select_steering_direction(
            torch.from_numpy(pos_by_layer[layer]).float(),
            torch.from_numpy(neg_by_layer[layer]).float(),
            "behavioral", behavioral_activations=beh_acts, behavioral_labels=behavioral_labels
        )
        layer_accuracies[layer] = acc

    val_size = int(len(test_ids) * val_split)
    val_ids = test_ids[:val_size]
    final_test_ids = test_ids[val_size:]
    print(f"\nAUTOTUNE: Validation={len(val_ids)}, Test={len(final_test_ids)}")

    task_name = list(all_pair_texts.values())[0].get("task_name", "truthfulqa_custom") if all_pair_texts else "truthfulqa_custom"
    evaluator = EvaluatorRotator(evaluator=None, task_name=task_name).current

    print(f"\nPhase 2: Searching for optimal steering configuration...")
    optimal_config = search_optimal_steering_config(
        adapter, list(train_ids), val_ids, all_pair_texts, evaluator,
        pos_by_layer, neg_by_layer, layer_accuracies, steering_vectors_dict, max_new_tokens
    )

    print(f"\nOptimal config found:")
    print(f"  Threshold: {optimal_config.accuracy_threshold:.2f}")
    print(f"  Base strength: {optimal_config.base_strength}")
    print(f"  Method: {optimal_config.steering_method}")
    print(f"  Val score: {optimal_config.val_score}/{len(val_ids)}")
    active = [(l, s) for l, s in optimal_config.layer_strengths.items() if s > 0]
    print(f"  Active layers: {len(active)}")

    scales_dict = {f"layer.{l}": s for l, s in optimal_config.layer_strengths.items()}
    methods_dict = {}
    for layer, strength in optimal_config.layer_strengths.items():
        if strength > 0:
            method = create_steering_method(optimal_config.steering_method, strength,
                pos_by_layer[layer], neg_by_layer[layer])
            if method:
                methods_dict[f"layer.{layer}"] = method

    return optimal_config, steering_vectors_dict, methods_dict, scales_dict, final_test_ids



"""Extracted TextAdapter methods: generation and chat template formatting."""

from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Dict, List, Optional

import torch

from wisent.core.primitives.model_interface.adapters.base import SteeringConfig
from wisent.core.primitives.model_interface.core.activations.core.atoms import LayerActivations
from wisent.core.primitives.models.modalities import TextContent
from wisent.core.primitives.models.config import get_generate_kwargs


class TextAdapterGenerationMixin:
    """Mixin providing generation and chat template methods for TextAdapter."""

    def forward_with_steering(
        self,
        content: TextContent | str,
        steering_vectors: LayerActivations,
        config: SteeringConfig | None = None,
    ) -> str:
        """
        Generate text with steering applied.

        Args:
            content: Input text
            steering_vectors: Steering vectors to apply
            config: Steering configuration

        Returns:
            Generated text with steering
        """
        if isinstance(content, str):
            content = TextContent(text=content)

        config = config or SteeringConfig()

        # Tokenize input
        inputs = self.tokenizer(
            content.text,
            return_tensors="pt",
            padding=True,
            truncation=True,
        )
        inputs = {k: v.to(self.model.device) for k, v in inputs.items()}

        # Apply steering during generation
        gen_kwargs = get_generate_kwargs()
        gen_kwargs["pad_token_id"] = self.tokenizer.pad_token_id
        with self._steering_hooks(steering_vectors, config):
            outputs = self.model.generate(
                **inputs,
                **gen_kwargs,
            )

        # Decode output
        generated = self.tokenizer.decode(outputs[0], skip_special_tokens=True)

        # Remove input prompt from output if present
        if generated.startswith(content.text):
            generated = generated[len(content.text):].strip()

        return generated

    def _generate_unsteered(
        self,
        content: TextContent | str,
        **kwargs: Any,
    ) -> str:
        """Generate text without steering."""
        if isinstance(content, str):
            content = TextContent(text=content)

        inputs = self.tokenizer(
            content.text,
            return_tensors="pt",
            padding=True,
            truncation=True,
        )
        inputs = {k: v.to(self.model.device) for k, v in inputs.items()}

        gen_kwargs = get_generate_kwargs(**kwargs)
        gen_kwargs["pad_token_id"] = self.tokenizer.pad_token_id
        with torch.no_grad():
            outputs = self.model.generate(
                **inputs,
                **gen_kwargs,
            )

        generated = self.tokenizer.decode(outputs[0], skip_special_tokens=True)

        if generated.startswith(content.text):
            generated = generated[len(content.text):].strip()

        return generated

    def apply_chat_template(
        self,
        messages: List[Dict[str, str]],
        add_generation_prompt: bool = True,
    ) -> str:
        """
        Apply chat template to messages.

        Args:
            messages: List of {"role": ..., "content": ...} dicts
            add_generation_prompt: Whether to add assistant prompt

        Returns:
            Formatted prompt string
        """
        return self.tokenizer.apply_chat_template(
            messages,
            tokenize=False,
            add_generation_prompt=add_generation_prompt,
        )



"""Null baseline comparison tests for geometry validation."""

from .effective_dim_null import (
    compute_null_effective_dimensions,
    compute_effective_dimensions_vs_null,
)
from .geometry_null import (
    compute_cone_null,
    compute_sphere_null,
    compute_cluster_null,
    compute_translation_null,
    compute_geometry_vs_null,
)

__all__ = [
    # Effective dimension null tests
    "compute_null_effective_dimensions",
    "compute_effective_dimensions_vs_null",
    # Geometry null tests
    "compute_cone_null",
    "compute_sphere_null",
    "compute_cluster_null",
    "compute_translation_null",
    "compute_geometry_vs_null",
]



"""Pipeline and objective function for steering optimization."""
import argparse
import json
import os
import time
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Optional

from wisent.core.utils.cli.optimize_steering.method_configs import (
    MethodConfig, CAAConfig, OstrzeConfig, MLPConfig, TECZAConfig,
    TETNOConfig, GROMConfig, NurtConfig, SzlakConfig, WicherConfig,
)
from wisent.core.utils.cli.optimize_steering.transport.method_configs_transport import PrzelomConfig
from wisent.core.utils.cli.optimize_steering.data.contrastive_pairs_data import execute_generate_pairs_from_task
from wisent.core.utils.cli.optimize_steering.data.activations_data import execute_get_activations
from wisent.core.utils.cli.optimize_steering.steering_objects import execute_create_steering_object
from wisent.core.utils.cli.optimize_steering.data.responses import execute_generate_responses
from wisent.core.utils.cli.optimize_steering.scores import execute_evaluate_responses
from wisent.core.utils.config_tools.constants import (
    SCORE_RANGE_MIN, GENERATION_DEFAULT_MAX_NEW_TOKENS,
    GENERATION_DEFAULT_TEMPERATURE, GENERATION_DEFAULT_TOP_P,
    EVAL_F1_THRESHOLD, EVAL_GENERATION_EMBEDDING_WEIGHT,
    EVAL_GENERATION_NLI_WEIGHT, SPLIT_RATIO_TRAIN_DEFAULT,
    METHODS_REQUIRING_QK_CAPTURE,
)
from wisent.core.control.steering_methods.configs.optimal import get_optimal, get_optimal_extraction_strategy


@dataclass
class OptimizationResult:
    """Result of a single optimization trial."""
    config: MethodConfig
    score: float
    details: dict


def _make_args(**kwargs):
    """Create an argparse.Namespace from kwargs."""
    args = argparse.Namespace()
    for k, v in kwargs.items():
        setattr(args, k, v)
    return args


def run_pipeline(
    model: str,
    task: str,
    config: MethodConfig,
    work_dir: str,
    strength: float,
    limit: Optional[int] = None,
    device: Optional[str] = None,
    enriched_pairs_file: Optional[str] = None,
    train_pairs_file: Optional[str] = None,
    test_pairs_file: Optional[str] = None,
    cached_model=None,
) -> OptimizationResult:
    """Run the full optimization pipeline for a single configuration.

    When train_pairs_file and test_pairs_file are set, the steering object
    is trained on train pairs and evaluated on test pairs (no data leakage).
    """
    activations_file = os.path.join(work_dir, "activations.json")
    steering_file = os.path.join(work_dir, "steering.pt")
    responses_file = os.path.join(work_dir, "responses.json")
    scores_file = os.path.join(work_dir, "scores.json")
    _ts = lambda: datetime.now(timezone.utc).isoformat()
    print(
        f"[run_pipeline] {_ts()} start: model={model}, task={task}, "
        f"method={config.method}, strength={strength}, limit={limit}",
        flush=True,
    )
    pipeline_t0 = time.monotonic()

    if enriched_pairs_file:
        activations_file = enriched_pairs_file
        eval_pairs_file = enriched_pairs_file
        eval_limit = limit
    elif train_pairs_file:
        layer = getattr(config, 'layer', None) or getattr(config, 'sensor_layer', None)
        if layer is None:
            raise ValueError("Config must specify 'layer' or 'sensor_layer'")
        needs_qk = config.method in METHODS_REQUIRING_QK_CAPTURE
        cached = None
        if not needs_qk:
            from wisent.core.reading.modules.utilities.data.enriched_builder import (
                build_enriched_from_hf, build_enriched_from_db,
            )
            cached = build_enriched_from_hf(
                model, task, layer, config.extraction_strategy, work_dir,
                train_pairs_file=train_pairs_file, limit=limit)
            if not cached:
                cached = build_enriched_from_db(
                    model, task, work_dir, config.extraction_strategy, limit=limit)
        if cached:
            activations_file = cached
        else:
            execute_get_activations(_make_args(
                pairs_file=train_pairs_file, model=model, output=activations_file,
                layers=str(layer), extraction_strategy=config.extraction_strategy,
                device=device, verbose=False, timing=False, raw=False,
                cached_model=cached_model,
            ))
        eval_pairs_file = test_pairs_file or train_pairs_file
        with open(eval_pairs_file) as f:
            eval_limit = len(json.load(f).get("pairs", []))
    else:
        pairs_file = os.path.join(work_dir, "pairs.json")
        execute_generate_pairs_from_task(_make_args(
            task_name=task, output=pairs_file, limit=limit, verbose=False,
            train_ratio=SPLIT_RATIO_TRAIN_DEFAULT,
        ))
        layer = getattr(config, 'layer', None) or getattr(config, 'sensor_layer', None)
        if layer is None:
            raise ValueError("Config must specify 'layer' or 'sensor_layer'")
        execute_get_activations(_make_args(
            pairs_file=pairs_file, model=model, output=activations_file,
            layers=str(layer), extraction_strategy=config.extraction_strategy,
            device=device, verbose=False, timing=False, raw=False,
            cached_model=cached_model,
        ))
        eval_pairs_file = pairs_file
        eval_limit = limit

    data_elapsed = time.monotonic() - pipeline_t0
    print(f"[run_pipeline] {_ts()} data prep done in {data_elapsed:.1f}s", flush=True)

    steer_t0 = time.monotonic()
    method_args = config.to_args()
    execute_create_steering_object(_make_args(
        enriched_pairs_file=activations_file, output=steering_file,
        verbose=False, timing=False, **method_args,
    ))
    steer_elapsed = time.monotonic() - steer_t0
    print(f"[run_pipeline] {_ts()} steering object created in {steer_elapsed:.1f}s", flush=True)

    steering_strategy = getattr(config, 'steering_strategy', get_optimal("steering_strategy"))
    gen_t0 = time.monotonic()
    print(
        f"[run_pipeline] {_ts()} generate start: eval_limit={eval_limit}, "
        f"strength={strength}, strategy={steering_strategy}, "
        f"max_new_tokens={GENERATION_DEFAULT_MAX_NEW_TOKENS}",
        flush=True,
    )
    execute_generate_responses(_make_args(
        task=task, input_file=eval_pairs_file, model=model, output=responses_file,
        num_questions=eval_limit, min_load_limit_questions=eval_limit,
        steering_object=steering_file, steering_strength=strength,
        steering_strategy=steering_strategy, use_steering=True,
        device=device, verbose=False, cached_model=cached_model,
        max_new_tokens=GENERATION_DEFAULT_MAX_NEW_TOKENS,
        temperature=GENERATION_DEFAULT_TEMPERATURE, top_p=GENERATION_DEFAULT_TOP_P,
    ))
    gen_elapsed = time.monotonic() - gen_t0
    print(f"[run_pipeline] {_ts()} generate done in {gen_elapsed:.1f}s", flush=True)

    eval_t0 = time.monotonic()
    execute_evaluate_responses(_make_args(
        input=responses_file, output=scores_file, task=task, verbose=False,
        f1_threshold=EVAL_F1_THRESHOLD,
        generation_embedding_weight=EVAL_GENERATION_EMBEDDING_WEIGHT,
        generation_nli_weight=EVAL_GENERATION_NLI_WEIGHT,
        train_ratio=SPLIT_RATIO_TRAIN_DEFAULT,
    ))
    eval_elapsed = time.monotonic() - eval_t0
    print(f"[run_pipeline] {_ts()} evaluate done in {eval_elapsed:.1f}s", flush=True)

    with open(scores_file) as f:
        scores_data = json.load(f)

    score = (
        scores_data.get("aggregated_metrics", {}).get("acc")
        or scores_data.get("accuracy")
        or scores_data.get("acc")
        or SCORE_RANGE_MIN
    )

    total_elapsed = time.monotonic() - pipeline_t0
    print(
        f"[run_pipeline] {_ts()} complete: score={score}, "
        f"total={total_elapsed:.1f}s (data={data_elapsed:.1f}s, "
        f"steer={steer_elapsed:.1f}s, gen={gen_elapsed:.1f}s, "
        f"eval={eval_elapsed:.1f}s)",
        flush=True,
    )
    return OptimizationResult(config=config, score=score, details=scores_data)


from wisent.core.utils.config_tools.constants import COMBO_OFFSET as _LAYER_OFFSET

_STRUCTURAL = frozenset({
    "layer", "sensor_layer", "steering_start", "steering_end",
    "strength", "steering_strategy", "direction_weighting",
    "extraction_component",
})
_NAME_MAP = {
    "tecza": {"min_cosine_similarity": "min_cosine_sim", "max_cosine_similarity": "max_cosine_sim"},
    "nurt": {"flow_hidden_dim": "hidden_dim"},
}
_ALREADY_PREFIXED = {"mlp": frozenset({"mlp_input_divisor", "mlp_early_stopping_patience"})}


def _prefix_params(method: str, params: dict) -> dict:
    """Auto-prefix non-structural params with the method name."""
    prefix = method.lower()
    overrides = _NAME_MAP.get(prefix, {})
    already = _ALREADY_PREFIXED.get(prefix, frozenset())
    extra = {}
    for k, v in params.items():
        if k in _STRUCTURAL:
            continue
        if k in already:
            extra[k] = v
            continue
        extra[f"{prefix}_{overrides.get(k, k)}"] = v
    return extra


def _build_config(method: str, params: dict) -> tuple[MethodConfig, float]:
    """Build a MethodConfig from flat params dict. Returns (config, strength)."""
    strength = params.get("strength", SCORE_RANGE_MIN)
    ext = get_optimal_extraction_strategy()
    steer = params.get("steering_strategy", get_optimal("steering_strategy"))
    m = method.upper()
    extra = _prefix_params(method, params)
    ec = params.get("extraction_component", get_optimal("extraction_component"))
    extra["extraction_component"] = ec
    kw = dict(extraction_strategy=ext, steering_strategy=steer, _extra_args=extra)
    if m in ("CAA", "OSTRZE", "MLP", "TECZA", "NURT", "SZLAK", "WICHER", "PRZELOM"):
        cls = {"CAA": CAAConfig, "OSTRZE": OstrzeConfig, "MLP": MLPConfig,
               "TECZA": TECZAConfig, "NURT": NurtConfig, "SZLAK": SzlakConfig,
               "WICHER": WicherConfig, "PRZELOM": PrzelomConfig}[m]
        method_name = {"OSTRZE": "Ostrze", "NURT": "nurt", "SZLAK": "szlak",
                       "WICHER": "wicher", "PRZELOM": "przelom"}.get(m, m)
        cfg = cls(method=method_name, layer=int(params["layer"]), **kw)
    elif m in ("TETNO", "GROM"):
        start, end = int(params["steering_start"]), int(params["steering_end"])
        if end < start:
            start, end = end, start
        layers = list(range(start, end + _LAYER_OFFSET))
        cls = TETNOConfig if m == "TETNO" else GROMConfig
        cfg = cls(method=m, sensor_layer=int(params["sensor_layer"]),
                  steering_layers=layers, **kw)
    else:
        raise ValueError(f"Unknown method: {method}")
    return cfg, strength


def create_objective(
    method: str,
    model: str,
    task: str,
    num_layers: int,
    limit: Optional[int],
    device: Optional[str],
    work_dir: str,
    enriched_pairs_file: Optional[str] = None,
    train_pairs_file: Optional[str] = None,
    test_pairs_file: Optional[str] = None,
    cached_model=None,
):
    """Create an objective function for the BaseOptimizer.

    Returns a callable that takes a flat params dict and returns a score.
    When train_pairs_file/test_pairs_file are set, trains on train and
    evaluates on test (no data leakage).
    """
    def objective(params: dict) -> float:
        config, strength = _build_config(method, params)
        return run_pipeline(
            model=model, task=task, config=config, work_dir=work_dir,
            strength=strength, limit=limit, device=device,
            enriched_pairs_file=enriched_pairs_file,
            train_pairs_file=train_pairs_file,
            test_pairs_file=test_pairs_file,
            cached_model=cached_model,
        ).score
    return objective



"""Auxiliary result type dataclasses for the geometry runner."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Any, Tuple
import torch
import numpy as np


@dataclass
class ZwiadResult:
    """Complete result from run_full_zwiad()."""
    benchmark: str
    model_name: str
    layer: int
    # Pair optimization
    optimal_n_pairs: int
    pair_search_history: List[Dict[str, Any]]
    saturation_status: str  # "OPTIMAL", "MAX_REACHED", "NO_SIGNAL"
    # Signal classification
    signal_exists: bool
    signal_type: str  # "LINEAR", "NONLINEAR", "MULTIMODAL", "NO_SIGNAL"
    # Method recommendation
    recommended_method: str  # "CAA", "TECZA", "TETNO", "GROM", "NO_METHOD"
    confidence: float
    reason: str
    # Core metrics
    linear_probe_accuracy: float
    best_nonlinear_accuracy: float
    knn_umap_accuracy: float
    knn_pacmap_accuracy: float
    # Geometry metrics (diff vectors)
    icd: float
    icd_top1_variance: float
    diff_mean_alignment: float
    caa_probe_alignment: float
    steerability_score: float
    effective_steering_dims: int
    # Multi-direction metrics
    multi_dir_gain: float
    spherical_silhouette_k2: float
    cluster_direction_angle: float
    # Quality metrics
    cohens_d: float
    bootstrap_std: float
    direction_stability: float
    # Steering vector (if signal exists)
    steering_direction: Optional[torch.Tensor] = None

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary (excluding tensor)."""
        return {
            "benchmark": self.benchmark,
            "model_name": self.model_name,
            "layer": self.layer,
            "optimal_n_pairs": self.optimal_n_pairs,
            "pair_search_history": self.pair_search_history,
            "saturation_status": self.saturation_status,
            "signal_exists": self.signal_exists,
            "signal_type": self.signal_type,
            "recommended_method": self.recommended_method,
            "confidence": self.confidence,
            "reason": self.reason,
            "linear_probe_accuracy": self.linear_probe_accuracy,
            "best_nonlinear_accuracy": self.best_nonlinear_accuracy,
            "knn_umap_accuracy": self.knn_umap_accuracy,
            "knn_pacmap_accuracy": self.knn_pacmap_accuracy,
            "icd": self.icd,
            "icd_top1_variance": self.icd_top1_variance,
            "diff_mean_alignment": self.diff_mean_alignment,
            "caa_probe_alignment": self.caa_probe_alignment,
            "steerability_score": self.steerability_score,
            "effective_steering_dims": self.effective_steering_dims,
            "multi_dir_gain": self.multi_dir_gain,
            "spherical_silhouette_k2": self.spherical_silhouette_k2,
            "cluster_direction_angle": self.cluster_direction_angle,
            "cohens_d": self.cohens_d,
            "bootstrap_std": self.bootstrap_std,
            "direction_stability": self.direction_stability,
        }


@dataclass
class ZwiadLayerResult:
    """Result from layer search - extends ZwiadResult with layer analysis."""
    best_result: ZwiadResult
    optimal_layer: int
    optimal_layer_range: List[int]
    layer_search_history: List[Dict[str, Any]]
    num_layers: int
    layers_tested: List[int]

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        d = self.best_result.to_dict()
        d.update({
            "optimal_layer": self.optimal_layer,
            "optimal_layer_range": self.optimal_layer_range,
            "layer_search_history": self.layer_search_history,
            "num_layers": self.num_layers,
            "layers_tested": self.layers_tested,
        })
        return d


@dataclass
class ComponentAnalysisResult:
    """Result from analyzing different transformer components."""
    best_component: str
    best_component_accuracy: float
    residual_accuracy: float
    mlp_accuracy: float
    attn_accuracy: float
    head_accuracies: Dict[int, float]
    top_heads: List[Tuple[int, float]]

    def to_dict(self) -> Dict[str, Any]:
        return {
            "best_component": self.best_component,
            "best_component_accuracy": self.best_component_accuracy,
            "residual_accuracy": self.residual_accuracy,
            "mlp_accuracy": self.mlp_accuracy,
            "attn_accuracy": self.attn_accuracy,
            "head_accuracies": self.head_accuracies,
            "top_heads": self.top_heads,
        }


@dataclass
class MultiConceptAnalysis:
    """Result from analyzing if multiple concepts exist in a contrastive pair set."""
    num_concepts_detected: int
    is_multi_concept: bool
    confidence: float
    icd: float
    icd_suggests_multi: bool
    cluster_count: int
    cluster_silhouette: float
    clusters_suggest_multi: bool
    pca_variance_ratio: List[float]
    pca_effective_rank: float
    pca_suggests_multi: bool
    multi_dir_accuracy: Dict[int, float]
    multi_dir_gain: float
    directions_suggest_multi: bool
    concept_directions: Optional[List[torch.Tensor]] = None
    concept_sizes: Optional[List[int]] = None
    concept_accuracies: Optional[List[float]] = None

    def to_dict(self) -> Dict[str, Any]:
        return {
            "num_concepts_detected": self.num_concepts_detected,
            "is_multi_concept": self.is_multi_concept,
            "confidence": self.confidence,
            "icd": self.icd,
            "icd_suggests_multi": self.icd_suggests_multi,
            "cluster_count": self.cluster_count,
            "cluster_silhouette": self.cluster_silhouette,
            "clusters_suggest_multi": self.clusters_suggest_multi,
            "pca_variance_ratio": self.pca_variance_ratio,
            "pca_effective_rank": self.pca_effective_rank,
            "pca_suggests_multi": self.pca_suggests_multi,
            "multi_dir_accuracy": self.multi_dir_accuracy,
            "multi_dir_gain": self.multi_dir_gain,
            "directions_suggest_multi": self.directions_suggest_multi,
            "concept_sizes": self.concept_sizes,
            "concept_accuracies": self.concept_accuracies,
        }


@dataclass
class ConceptValidityResult:
    """Result from analyzing if a set of pairs represents a valid concept."""
    is_valid_concept: bool
    validity_score: float
    concept_level: str  # "instance", "category", "domain", "noise"
    coherence_score: float
    coherence_std: float
    is_coherent: bool
    stability_score: float
    stability_std: float
    is_stable: bool
    signal_strength: float
    signal_to_noise: float
    has_signal: bool
    icd: float
    specificity: float

    def to_dict(self) -> Dict[str, Any]:
        return {
            "is_valid_concept": self.is_valid_concept,
            "validity_score": self.validity_score,
            "concept_level": self.concept_level,
            "coherence_score": self.coherence_score,
            "coherence_std": self.coherence_std,
            "is_coherent": self.is_coherent,
            "stability_score": self.stability_score,
            "stability_std": self.stability_std,
            "is_stable": self.is_stable,
            "signal_strength": self.signal_strength,
            "signal_to_noise": self.signal_to_noise,
            "has_signal": self.has_signal,
            "icd": self.icd,
            "specificity": self.specificity,
        }


@dataclass
class ConceptDecomposition:
    """Result from decomposing a contrastive pair set into constituent concepts."""
    n_concepts: int
    decomposition_method: str  # "clustering", "nmf", "ica"
    concept_directions: List[torch.Tensor]
    concept_sizes: List[int]
    concept_coherences: List[float]
    concept_validities: List[bool]
    concept_relationships: List[Dict[str, Any]]
    pair_to_concept: List[int]
    pair_concept_scores: np.ndarray
    total_explained_variance: float
    reconstruction_error: float

    def to_dict(self) -> Dict[str, Any]:
        return {
            "n_concepts": self.n_concepts,
            "decomposition_method": self.decomposition_method,
            "concept_sizes": self.concept_sizes,
            "concept_coherences": self.concept_coherences,
            "concept_validities": self.concept_validities,
            "concept_relationships": self.concept_relationships,
            "pair_to_concept": self.pair_to_concept,
            "total_explained_variance": self.total_explained_variance,
            "reconstruction_error": self.reconstruction_error,
        }



"""Wizard recommendation text for each subgoal."""

RECOMMENDATIONS = {
    "Generate synthetic contrastive pairs from a custom trait": (
        "generate-pairs",
        "Generates contrastive pairs where the model answers the same "
        "prompts with and without a specified trait.",
    ),
    "Generate pairs from an lm-eval benchmark task": (
        "generate-pairs-from-task",
        "Extracts contrastive pairs from an lm-eval harness task.",
    ),
    "Generate model responses to evaluation questions": (
        "generate-responses",
        "Generates model responses, optionally with steering applied.",
    ),
    "Run the full synthetic pipeline end-to-end": (
        "synthetic",
        "Runs the complete synthetic pipeline: prompts, responses, "
        "and enriched pairs.",
    ),
    "From an lm-eval task (full pipeline)": (
        "generate-vector-from-task",
        "Full pipeline: extracts pairs, collects activations, and "
        "trains a steering vector from a benchmark.",
    ),
    "From synthetic contrastive pairs": (
        "generate-vector-from-synthetic",
        "Full pipeline from synthetic data: generates pairs, collects "
        "activations, and trains a steering vector.",
    ),
    "From existing enriched pairs": (
        "create-steering-vector",
        "Creates a steering vector from already-enriched contrastive "
        "pairs with activations.",
    ),
    "Discover the best steering direction automatically": (
        "discover-steering",
        "Searches for the optimal steering direction by trying multiple "
        "methods, layers, and hyperparameters.",
    ),
    "Combine multiple steering vectors at inference": (
        "multi-steer",
        "Applies multiple steering vectors simultaneously during "
        "generation with individual strengths.",
    ),
    "Visualize how steering affects activation space": (
        "steering-viz",
        "Creates PCA/UMAP plots comparing steered vs unsteered "
        "activations.",
    ),
    "Verify steered activations are aligned correctly": (
        "verify-steering",
        "Verifies that a steered model produces activations aligned "
        "with the intended direction.",
    ),
    "Compare steering objects across traits": (
        "compare-steering",
        "Compares multiple steering objects: cosine similarity, overlap, "
        "and interference between directions.",
    ),
    "Evaluate response quality with embedded evaluator": (
        "evaluate-responses",
        "Evaluates generated responses using NLI, embedding similarity, "
        "or other built-in evaluators.",
    ),
    "Evaluate model refusal rate on harmful prompts": (
        "evaluate-refusal",
        "Measures how often a model refuses to answer harmful prompts.",
    ),
    "Run all optimizations at once": (
        "optimize",
        "Runs the complete optimization suite: classification, steering, "
        "and evaluation.",
    ),
    "Optimize classification thresholds": (
        "optimize-classification",
        "Tunes MLP classifier hyperparameters for separating positive "
        "and negative activations.",
    ),
    "Optimize steering parameters (method, layer, strength)": (
        "optimize-steering",
        "Searches for the best method, target layer, and strength via "
        "Bayesian optimization.",
    ),
    "Find optimal training sample size": (
        "optimize-sample-size",
        "Determines the minimum number of contrastive pairs needed.",
    ),
    "Optimize weight modification parameters": (
        "optimize-weights",
        "Optimizes parameters for permanent weight modification.",
    ),
    "Manage cached optimization results": (
        "optimization-cache",
        "View, clear, or export cached optimization results.",
    ),
    "Find the best steering method for a benchmark": (
        "find-best-method",
        "Trials every steering method on a benchmark and ranks them.",
    ),
    "Diagnose contrastive pair quality": (
        "diagnose-pairs",
        "Analyzes pairs for quality issues: semantic divergence, length "
        "imbalance, duplication, and consistency.",
    ),
    "Diagnose steering vector quality": (
        "diagnose-vectors",
        "Analyzes a trained steering vector: separability, stability, "
        "and interference.",
    ),
    "Check if a representation is linear": (
        "check-linearity",
        "Tests whether a concept is represented linearly in model "
        "activations.",
    ),
    "Cluster benchmarks by direction similarity": (
        "cluster-benchmarks",
        "Groups benchmarks whose steering directions are similar.",
    ),
    "Search for unified goodness direction": (
        "geometry-search",
        "Searches for a single direction that improves performance "
        "across multiple benchmarks.",
    ),
    "Run full Zwiad geometry analysis": (
        "zwiad",
        "Runs comprehensive geometry analysis: layer sensitivity, "
        "method comparison, and similarity heatmaps.",
    ),
    "Permanently modify model weights with steering": (
        "modify-weights",
        "Permanently applies steering to model weights and saves "
        "a new model.",
    ),
    "Collect activations from contrastive pairs": (
        "get-activations",
        "Extracts hidden-state activations from a model for each "
        "contrastive pair.",
    ),
    "Train a unified goodness vector from benchmarks": (
        "train-unified-goodness",
        "Trains a single steering vector from multiple benchmark "
        "tasks capturing a general goodness direction.",
    ),
    "View and update inference settings": (
        "inference-config",
        "View or modify inference configuration: temperature, top-p, "
        "max tokens, and generation parameters.",
    ),
    "Run evaluation tasks": (
        "tasks",
        "Lists or runs evaluation tasks from the lm-eval harness.",
    ),
}



"""Geometry, linearity, stability, and subspace constants."""
from wisent.core.utils.config_tools.constants.for_experiments.by_domain.analysis._analysis import *  # noqa: F401,F403

# --- Geometry metric constants (A: fixed by definition, B: data-derivable) ---
# Category A reuse validated constants (CHANCE_LEVEL_ACCURACY, SCORE_RANGE_MIN, COMBO_OFFSET)
# Category C (non-derivable design choices) are now required CLI arguments

GEOMETRY_VARIANCE_EXPLAINED_90PCT = 0.9  # "90%" in "dims for 90% variance"

# Category B: min_cloud_points (minimum for centroid + spread, need >= 3 for std)
GEOMETRY_MIN_CLOUD_POINTS = 3

# Category B: CV folds standard (maximum)
GEOMETRY_CV_FOLDS_MAX = 5
GEOMETRY_CV_FOLDS_MIN = 2

# Category B: KNN heuristic bounds
GEOMETRY_KNN_K_MIN = 3
GEOMETRY_KNN_K_MAX = 15

# Category B: PCA null model cap
GEOMETRY_PCA_NULL_CAP = 50

# Category B: probe validation floor
GEOMETRY_PROBE_VALIDATION_MIN = 0.1
GEOMETRY_PROBE_VALIDATION_MIN_SAMPLES = 2

# Category B: classifier self-configuration bounds (data-adaptive)
CLASSIFIER_HIDDEN_DIM_MIN = 32
CLASSIFIER_HIDDEN_DIM_MAX = 512
CLASSIFIER_BATCH_SIZE_MIN = 8
CLASSIFIER_BATCH_SIZE_MAX = 64
CLASSIFIER_BATCH_DIVISOR = 4
CLASSIFIER_TEST_SIZE_MIN = 0.1
CLASSIFIER_TEST_SIZE_MAX = 0.3
CLASSIFIER_MIN_TEST_SAMPLES = 10
CLASSIFIER_DROPOUT_MIN = 0.05
CLASSIFIER_DROPOUT_MAX = 0.5
CLASSIFIER_LOG_DIVISOR = 10
CLASSIFIER_DEFAULT_LR = 0.001



"""
Run lm-evaluation-harness code evaluation tasks in Docker (sandboxed).

Supported tasks: any, in particular:
    - humaneval, humaneval_instruct, humaneval_plus, humaneval_plus_instruct
    - mbpp, mbpp_instruct, mbpp_plus, mbpp_plus_instruct

Args:
    --tasks       lm_eval task(s), comma-separated (required)
    --model       HuggingFace model name (required)
    --batch_size  Batch size (default: 4)
    --device      Device to use (required)
    --output_dir  Output directory (default: ./output)
"""

import argparse
import subprocess
import os
from pathlib import Path
from wisent.core.utils.infra_tools.infra.core.hardware import eval_batch_size


IMAGE_NAME = "lm-eval:code-eval"
SCRIPT_DIR = Path(__file__).parent


def build_image() -> None:
    """Build Docker image if it doesn't exist."""
    result = subprocess.run(
        ["docker", "image", "inspect", IMAGE_NAME],
        capture_output=True,
    )
    if result.returncode != 0:
        print(f"Building Docker image: {IMAGE_NAME}")
        subprocess.run(
            ["docker", "build", "-t", IMAGE_NAME, str(SCRIPT_DIR)],
            check=True,
        )


def run_code_eval(
    model: str,
    tasks: str,
    device: str,
    batch_size: int,
    output_dir: Path,
) -> None:
    """Run lm_eval code evaluation tasks in sandboxed Docker container."""
    build_image()

    # Setup directories
    output_dir.mkdir(parents=True, exist_ok=True)

    hf_cache = Path(os.environ.get("HF_HOME", Path.home() / ".cache" / "huggingface"))
    hf_cache.mkdir(parents=True, exist_ok=True)

    # Build lm_eval arguments
    lm_eval_args = [
        "--model", "hf",
        "--model_args", f"pretrained={model}",
        "--tasks", tasks,
        "--device", device,
        "--batch_size", str(batch_size),
        "--confirm_run_unsafe_code",
        "--output_path", "/home/sandbox/output",
    ]


    # Docker command
    docker_cmd = [
        "docker", "run", "--rm", "-it",
        "--gpus", "all",
        "--user", f"{os.getuid()}:{os.getgid()}",
        "--cap-drop=ALL",
        "--security-opt=no-new-privileges",
        "-v", f"{hf_cache}:/home/sandbox/.cache/huggingface",
        "-v", f"{output_dir}:/home/sandbox/output",
        "-e", "HF_ALLOW_CODE_EVAL=1",
        "-e", "HF_HOME=/home/sandbox/.cache/huggingface",
        IMAGE_NAME,
        *lm_eval_args,
    ]

    print("Running lm_eval in Docker...")
    print(f"  Image: {IMAGE_NAME}")
    print(f"  Model: {model}")
    print(f"  Tasks: {tasks}")
    print(f"  Output: {output_dir}")
    print()

    subprocess.run(docker_cmd, check=True)

    print()
    print(f"Results saved to: {output_dir}")


def main():
    parser = argparse.ArgumentParser(description="Run code evaluation tasks in Docker")
    parser.add_argument("--model", required=True, help="HuggingFace model name")
    parser.add_argument("--tasks", required=True, help="lm_eval task(s), comma-separated")
    parser.add_argument("--device", required=True, help="Device to use")
    parser.add_argument("--batch_size", type=int, default=None, help="Batch size (auto-detected if omitted)")
    parser.add_argument("--output_dir", type=Path, default=None, help="Output directory")

    args = parser.parse_args()

    batch_size = args.batch_size if args.batch_size is not None else eval_batch_size()
    output_dir = args.output_dir if args.output_dir is not None else SCRIPT_DIR / "output"

    run_code_eval(
        model=args.model,
        tasks=args.tasks,
        device=args.device,
        batch_size=batch_size,
        output_dir=output_dir,
    )


if __name__ == "__main__":
    main()



"""Map argparse actions to Gradio components and back.

Introspects argparse action types and creates matching Gradio form
elements. Also provides reverse mapping from Gradio values to CLI args.
"""

import argparse
import gradio as gr
from wisent.core.utils.config_tools.constants import (
    INDEX_FIRST,
    GRADIO_APPEND_LINES,
)


def action_to_component(action):
    """Convert an argparse action to a Gradio component.

    Returns (component, arg_name) or (None, None) for skip actions.
    """
    if isinstance(action, argparse._HelpAction):
        return None, None
    if isinstance(action, argparse._SubParsersAction):
        return None, None

    dest = action.dest
    if dest == "command":
        return None, None

    label = _build_label(action)
    default = action.default

    if isinstance(action, (argparse._StoreTrueAction, argparse._StoreFalseAction)):
        val = bool(default) if default is not None else False
        return gr.Checkbox(label=label, value=val), dest

    if action.choices is not None:
        choices_list = [str(c) for c in action.choices]
        val = str(default) if default is not None else None
        return gr.Dropdown(
            label=label, choices=choices_list, value=val,
            allow_custom_value=True,
        ), dest

    if action.type is int:
        val = default if isinstance(default, int) else None
        return gr.Number(label=label, value=val, precision=INDEX_FIRST), dest

    if action.type is float:
        val = default if isinstance(default, (int, float)) else None
        return gr.Number(label=label, value=val), dest

    if isinstance(action, argparse._AppendAction):
        return gr.Textbox(
            label=label, value=str(default) if default else "",
            lines=GRADIO_APPEND_LINES, placeholder="One value per line",
        ), dest

    if action.nargs in ("+", "*"):
        val = " ".join(str(v) for v in default) if default else ""
        return gr.Textbox(
            label=label, value=val,
            placeholder="Space-separated values",
        ), dest

    val = str(default) if default is not None else ""
    required = action.required if hasattr(action, "required") else False
    is_positional = not action.option_strings
    placeholder = "Required" if (required or is_positional) else ""
    return gr.Textbox(label=label, value=val, placeholder=placeholder), dest


def _build_label(action):
    """Build a human-readable label from an argparse action."""
    if action.option_strings:
        name = action.option_strings[INDEX_FIRST]
        for opt in action.option_strings:
            if opt.startswith("--"):
                name = opt
                break
        name = name.lstrip("-").replace("-", " ").replace("_", " ").title()
    else:
        name = action.dest.replace("_", " ").replace("-", " ").title()

    if action.help and action.help != argparse.SUPPRESS:
        return f"{name} -- {action.help}"
    return name


def components_to_args(component_values, component_dests, command_name):
    """Convert Gradio component values back to CLI argument list.

    Args:
        component_values: List of values from Gradio components.
        component_dests: List of (dest_name, action) tuples.
        command_name: The CLI command name.

    Returns:
        List of CLI argument strings.
    """
    args = []
    for value, (dest, action) in zip(component_values, component_dests):
        if value is None or value == "":
            continue

        if isinstance(action, (argparse._StoreTrueAction,)):
            if value:
                flag = _dest_to_flag(dest)
                args.append(flag)
            continue

        if isinstance(action, argparse._StoreFalseAction):
            if not value:
                flag = _dest_to_flag(dest)
                args.append(flag)
            continue

        if not action.option_strings:
            args.append(str(value))
            continue

        flag = _dest_to_flag(dest)

        if isinstance(action, argparse._AppendAction):
            for line in str(value).strip().splitlines():
                line = line.strip()
                if line:
                    args.extend([flag, line])
            continue

        if action.nargs in ("+", "*"):
            parts = str(value).strip().split()
            if parts:
                args.append(flag)
                args.extend(parts)
            continue

        args.extend([flag, str(value)])
    return args


def _dest_to_flag(dest):
    """Convert argparse dest back to CLI flag string."""
    return "--" + dest.replace("_", "-")



"""Evaluation and IO mixin for SteeringMethod."""
from __future__ import annotations

import torch
from typing import Any, Dict, Optional, TYPE_CHECKING
from wisent.core.primitives.contrastive_pairs import ContrastivePairSet
from wisent.core.utils.infra_tools.errors import NoTrainedVectorsError, InsufficientDataError

if TYPE_CHECKING:
    from wisent.core.control.steering_core.steering import SteeringType


class SteeringEvalIOMixin:
    """Mixin providing evaluation and IO methods."""

    def evaluate(self, contrastive_pair_set: ContrastivePairSet) -> Dict[str, Any]:
        """
        Evaluate the steering method on a ContrastivePairSet.

        Args:
            contrastive_pair_set: Set of contrastive pairs for evaluation

        Returns:
            Dictionary with evaluation metrics
        """
        if self.classifier is None:
            raise NoTrainedVectorsError()

        # Get positive and negative activations
        pos_activations, neg_activations = contrastive_pair_set.get_activation_pairs()

        # Predict on positive activations (should be low scores)
        pos_predictions = []
        for activation in pos_activations:
            pred = self.predict_proba(activation)
            pos_predictions.append(pred)

        # Predict on negative activations (should be high scores)
        neg_predictions = []
        for activation in neg_activations:
            pred = self.predict_proba(activation)
            neg_predictions.append(pred)

        # Calculate metrics
        # True Positives: negative activations correctly identified as harmful (pred >= threshold)
        true_positives = sum(1 for pred in neg_predictions if pred >= self.threshold)

        # False Positives: positive activations incorrectly identified as harmful (pred >= threshold)
        false_positives = sum(1 for pred in pos_predictions if pred >= self.threshold)

        # True Negatives: positive activations correctly identified as harmless (pred < threshold)
        true_negatives = sum(1 for pred in pos_predictions if pred < self.threshold)

        # False Negatives: negative activations incorrectly identified as harmless (pred < threshold)
        false_negatives = sum(1 for pred in neg_predictions if pred < self.threshold)

        # Calculate metrics
        detection_rate = true_positives / len(neg_predictions) if neg_predictions else 0
        false_positive_rate = false_positives / len(pos_predictions) if pos_predictions else 0

        precision = true_positives / (true_positives + false_positives) if (true_positives + false_positives) > 0 else 0
        recall = true_positives / (true_positives + false_negatives) if (true_positives + false_negatives) > 0 else 0
        f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0

        accuracy = (
            (true_positives + true_negatives) / (len(pos_predictions) + len(neg_predictions))
            if (pos_predictions or neg_predictions)
            else 0
        )

        return {
            "detection_rate": detection_rate,
            "false_positive_rate": false_positive_rate,
            "precision": precision,
            "recall": recall,
            "f1": f1,
            "accuracy": accuracy,
            "true_positives": true_positives,
            "false_positives": false_positives,
            "true_negatives": true_negatives,
            "false_negatives": false_negatives,
            "num_positive_samples": len(pos_predictions),
            "num_negative_samples": len(neg_predictions),
            "threshold": self.threshold,
        }

    def save_model(self, save_path: str) -> bool:
        """
        Save the steering method to disk.

        Args:
            save_path: Path to save the model

        Returns:
            Success flag
        """
        if self.classifier is None:
            return False

        try:
            self.classifier.save_model(save_path)
            return True
        except Exception:
            return False

    def load_model(self, model_path: str) -> bool:
        """
        Load a steering method from disk.

        Args:
            model_path: Path to the saved model

        Returns:
            Success flag
        """
        try:
            self.classifier = Classifier(
                model_type=self.method_type.value, device=self.device, threshold=self.threshold, model_path=model_path
            )
            return True
        except Exception:
            return False

    @classmethod
    def create_and_train(
        cls,
        method_type: SteeringType,
        contrastive_pair_set: ContrastivePairSet,
        device: Optional[str] = None,
        threshold: float,
        **training_kwargs,
    ) -> "SteeringMethod":
        """
        Create and train a SteeringMethod in one step.

        Args:
            method_type: Type of steering method
            contrastive_pair_set: Training data
            device: Device to use
            threshold: Classification threshold
            **training_kwargs: Additional training parameters

        Returns:
            Trained SteeringMethod
        """
        steering = cls(method_type=method_type, device=device, threshold=threshold)
        steering.train(contrastive_pair_set, **training_kwargs)
        return steering



"""CLI entry point for estimating unified goodness training time."""
from wisent.core.utils.config_tools.constants import SEPARATOR_WIDTH_WIDE, SEPARATOR_WIDTH_HALF
from wisent.core.utils.cli.analysis.training.estimate_time_functions import (
    BENCHMARK_SIZES, BENCHMARK_LOAD_TIMES, estimate_runtime,
    get_benchmark_load_time, get_benchmark_size, format_time,
)


def main():
    parser = argparse.ArgumentParser(
        description="Estimate runtime for train-unified-goodness command (ALL 327 benchmarks)"
    )
    parser.add_argument(
        "--max-benchmarks", type=int, default=None,
        help="Maximum number of benchmarks to use (default: ALL 327)"
    )
    parser.add_argument(
        "--cap-pairs-per-benchmark", type=int, default=None,
        help="Cap pairs per benchmark (benchmarks with more get randomly sampled down)"
    )
    parser.add_argument(
        "--train-ratio", type=float, default=None,
        help="Fraction of pairs for training (default: 0.8 = 80%% train, 20%% eval)"
    )
    parser.add_argument(
        "--eval-scales", type=int, required=True,
        help="Number of steering scales to evaluate"
    )
    parser.add_argument(
        "--skip-evaluation", action="store_true",
        help="Skip evaluation phase"
    )
    parser.add_argument(
        "--device", choices=["cuda", "cpu", "mps", "auto"], required=True,
        help="Device for computation"
    )
    parser.add_argument(
        "--show-breakdown", action="store_true",
        help="Show detailed time breakdown"
    )
    parser.add_argument(
        "--benchmark-fast-load-threshold", type=int, required=True,
        help="Load time threshold (seconds) to classify benchmarks as medium vs low priority"
    )
    parser.add_argument(
        "--show-benchmarks", action="store_true",
        help="List all benchmarks with their estimated sizes"
    )
    
    args = parser.parse_args()
    
    # Import benchmark list - use ALL benchmarks (same as test_all_benchmarks.py)
    try:
        from wisent.core.utils.cli.train_unified_goodness import load_all_benchmarks
        all_benchmark_names, broken = load_all_benchmarks()
    except ImportError:
        print("Error: Cannot import load_all_benchmarks. Run from wisent-open-source directory.")
        sys.exit(1)
    
    # Convert to dict format for estimate_runtime (name -> config with priority)
    # Since we don't have priority info from the JSON files, we'll estimate based on name
    selected = {}
    for name in all_benchmark_names:
        # Estimate priority based on common fast benchmarks
        if name in BENCHMARK_LOAD_TIMES:
            if BENCHMARK_LOAD_TIMES[name] < 15:
                priority = 'high'
            elif BENCHMARK_LOAD_TIMES[name] < args.benchmark_fast_load_threshold:
                priority = 'medium'
            else:
                priority = 'low'
        else:
            priority = 'medium'  # Default
        selected[name] = {'priority': priority}
    
    # Apply max limit
    if args.max_benchmarks and len(selected) > args.max_benchmarks:
        selected = dict(list(selected.items())[:args.max_benchmarks])
    
    # Print header
    print("\n" + "=" * SEPARATOR_WIDTH_WIDE)
    print("  UNIFIED GOODNESS TRAINING - TIME ESTIMATE")
    print("=" * SEPARATOR_WIDTH_WIDE)
    
    # Print configuration
    print(f"\n CONFIGURATION:")
    print(f"   Total available: {len(all_benchmark_names)} benchmarks")
    print(f"   Using: {len(selected)} benchmarks")
    if args.cap_pairs_per_benchmark:
        print(f"   Pairs cap: {args.cap_pairs_per_benchmark} per benchmark (random sampling)")
    else:
        print(f"   Pairs: ALL available (80% train / 20% eval)")
    print(f"   Train ratio: {args.train_ratio}")
    print(f"   Eval scales: {args.eval_scales}")
    print(f"   Skip evaluation: {args.skip_evaluation}")
    print(f"   Device: {args.device}")
    
    # Calculate estimate
    estimate = estimate_runtime(
        benchmarks=selected,
        num_eval_scales=args.eval_scales,
        train_ratio=args.train_ratio,
        skip_evaluation=args.skip_evaluation,
        device=args.device,
        cap_pairs_per_benchmark=args.cap_pairs_per_benchmark,
    )
    
    # Show benchmarks if requested
    if args.show_benchmarks:
        print(f"\n SELECTED BENCHMARKS ({len(selected)}):")
        print(f"   {'#':<4s} {'Benchmark':<30s} {'Priority':<8s} {'Pairs':>8s} {'Train':>8s} {'Eval':>8s}")
        print(f"   {'-'*4} {'-'*30} {'-'*8} {'-'*8} {'-'*8} {'-'*8}")
        for i, detail in enumerate(estimate['benchmark_details'], 1):
            print(f"   {i:<4d} {detail['name']:<30s} {detail['priority']:<8s} "
                  f"{detail['total_pairs']:>8d} {detail['train_pairs']:>8d} {detail['eval_pairs']:>8d}")
        print(f"   {'-'*4} {'-'*30} {'-'*8} {'-'*8} {'-'*8} {'-'*8}")
        print(f"   {'':4s} {'TOTAL':<30s} {'':8s} {estimate['total_pairs']:>8d} "
              f"{estimate['total_train_pairs']:>8d} {estimate['total_eval_pairs']:>8d}")
    
    # Print summary
    print(f"\n  ESTIMATED TOTAL TIME: {format_time(estimate['total_seconds'])}")
    print(f"   ({estimate['total_minutes']:.1f} minutes / {estimate['total_hours']:.2f} hours)")
    
    # Print breakdown
    if args.show_breakdown:
        print(f"\n TIME BREAKDOWN:")
        print(f"   Model loading:        {format_time(estimate['model_loading']):>10s}")
        print(f"   Benchmark loading:    {format_time(estimate['benchmark_loading']):>10s}")
        print(f"   Benchmark setup:      {format_time(estimate['benchmark_setup']):>10s}")
        print(f"   Pair generation:      {format_time(estimate['pair_generation']):>10s}")
        print(f"   Activation collection:{format_time(estimate['activation_collection']):>10s}")
        print(f"   Vector training:      {format_time(estimate['vector_training']):>10s}")
        print(f"   Evaluation:           {format_time(estimate['evaluation']):>10s}")
        print(f"   " + "-" * SEPARATOR_WIDTH_HALF)
        print(f"   TOTAL:                {format_time(estimate['total_seconds']):>10s}")
    
    # Print data summary
    print(f"\n DATA SUMMARY:")
    print(f"   Total pairs:          {estimate['total_pairs']:,}")
    print(f"   Training pairs (80%): {estimate['total_train_pairs']:,}")
    print(f"   Eval pairs (20%):     {estimate['total_eval_pairs']:,}")
    if not args.skip_evaluation:
        print(f"   Eval samples (x{args.eval_scales} scales): {estimate.get('total_eval_samples', 0):,}")
    
    # Print quick reference table with realistic sizes
    print(f"\n QUICK REFERENCE - OTHER CONFIGURATIONS:")
    print(f"   {'Config':<45s} {'Pairs':>10s} {'Time':>12s}")
    print(f"   {'-'*45} {'-'*10} {'-'*12}")
    
    # Build configs with subset of benchmarks (using full list now)
    def get_subset(max_n=None):
        subset = {}
        for name in all_benchmark_names:
            if name in BENCHMARK_LOAD_TIMES:
                if BENCHMARK_LOAD_TIMES[name] < 15:
                    priority = 'high'
                elif BENCHMARK_LOAD_TIMES[name] < args.benchmark_fast_load_threshold:
                    priority = 'medium'
                else:
                    priority = 'low'
            else:
                priority = 'medium'
            subset[name] = {'priority': priority}
        if max_n and len(subset) > max_n:
            subset = dict(list(subset.items())[:max_n])
        return subset
    
    configs = [
        ('10 benchmarks, skip eval', get_subset(10), True, None),
        ('50 benchmarks, full eval', get_subset(50), False, None),
        (f'ALL ({len(all_benchmark_names)}) benchmarks, no cap', get_subset(), False, None),
        (f'ALL ({len(all_benchmark_names)}) benchmarks, cap 10000', get_subset(), False, 10000),
        (f'ALL ({len(all_benchmark_names)}) benchmarks, cap 5000', get_subset(), False, 5000),
        (f'ALL ({len(all_benchmark_names)}) benchmarks, cap 1000', get_subset(), False, 1000),
    ]
    
    for name, bench_subset, skip_eval, limit in configs:
        est = estimate_runtime(
            benchmarks=bench_subset,
            num_eval_scales=args.eval_scales,
            train_ratio=args.train_ratio,
            skip_evaluation=skip_eval,
            device=args.device,
            cap_pairs_per_benchmark=limit,
        )
        print(f"   {name:<45s} {est['total_pairs']:>10,} {format_time(est['total_seconds']):>12s}")
    
    print("\n" + "=" * SEPARATOR_WIDTH_WIDE)

    return estimate


if __name__ == "__main__":
    main()



"""
Custom evaluator classes for user-defined evaluation functions.

This module provides flexible evaluation interfaces:
- CustomEvaluator: Base class for custom evaluators
- CallableEvaluator: Wrap any Python callable as an evaluator
- APIEvaluator: Base class for API-based evaluators (e.g., GPTZero)
"""

from __future__ import annotations

import importlib
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Protocol, Union

from wisent.core.utils.config_tools.constants import SCORE_RANGE_MIN, SCORE_RANGE_MAX
from wisent.core.reading.evaluators.core.atoms import BaseEvaluator, EvalResult

# Re-export from helpers
from wisent.core.reading.evaluators.custom._custom_evaluator_helpers import (
    create_custom_evaluator,
)

__all__ = [
    "CustomEvaluator",
    "CallableEvaluator",
    "APIEvaluator",
    "create_custom_evaluator",
    "EvaluatorProtocol",
]

logger = logging.getLogger(__name__)


class EvaluatorProtocol(Protocol):
    """Protocol for custom evaluator functions."""
    def __call__(self, response: str, **kwargs) -> Union[float, Dict[str, Any]]: ...


@dataclass
class CustomEvaluatorConfig:
    """Configuration for custom evaluators."""
    name: Optional[str] = None
    description: str = "Custom evaluator"
    invert_score: bool = False
    score_key: Optional[str] = None
    normalize: bool = True


class CustomEvaluator(ABC):
    """Abstract base class for custom evaluators."""

    def __init__(
        self,
        name: str,
        description: str,
        config: Optional[CustomEvaluatorConfig] = None,
    ):
        self.name = name
        self.description = description
        self.config = config or CustomEvaluatorConfig(name=name, description=description)

    @abstractmethod
    def evaluate_response(self, response: str, **kwargs) -> Union[float, Dict[str, Any]]:
        """Evaluate a single response."""
        pass

    def evaluate_batch(
        self, responses: List[str], **kwargs,
    ) -> List[Dict[str, Any]]:
        """Evaluate a batch of responses."""
        results = []
        for response in responses:
            result = self.evaluate_response(response, **kwargs)
            if isinstance(result, (int, float)):
                result = {"score": float(result)}
            results.append(result)
        return results

    def _normalize_result(self, result: Union[float, Dict[str, Any]]) -> Dict[str, Any]:
        """Normalize evaluation result to standard format."""
        if isinstance(result, (int, float)):
            score = float(result)
            result = {"score": score}
        else:
            score = result.get(self.config.score_key, result.get("score", 0.0))

        if self.config.normalize:
            score = (score - SCORE_RANGE_MIN) / (SCORE_RANGE_MAX - SCORE_RANGE_MIN)
            score = max(SCORE_RANGE_MIN, min(SCORE_RANGE_MAX, score))

        if self.config.invert_score:
            score = 1.0 - score

        result["score"] = score
        return result

    def __call__(self, response: str, **kwargs) -> Dict[str, Any]:
        """Make the evaluator callable."""
        result = self.evaluate_response(response, **kwargs)
        return self._normalize_result(result)


class CallableEvaluator(CustomEvaluator):
    """Wrap any callable as an evaluator."""

    def __init__(
        self,
        fn: Callable[[str], Union[float, Dict[str, Any]]],
        name: str,
        description: str = "Callable evaluator",
        config: Optional[CustomEvaluatorConfig] = None,
    ):
        super().__init__(name=name, description=description, config=config)
        self._fn = fn

    def evaluate_response(self, response: str, **kwargs) -> Union[float, Dict[str, Any]]:
        return self._fn(response, **kwargs) if kwargs else self._fn(response)


class APIEvaluator(CustomEvaluator):
    """Base class for API-based evaluators."""

    def __init__(
        self,
        name: str,
        description: str,
        api_timeout: int,
        max_retries: int,
        retry_delay: float,
        config: Optional[CustomEvaluatorConfig] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        cache_responses: bool = True,
    ):
        super().__init__(name=name, description=description, config=config)
        self.api_key = api_key
        self.base_url = base_url
        self.timeout = api_timeout
        self.max_retries = max_retries
        self.retry_delay = retry_delay
        self.cache_responses = cache_responses
        self._cache: Dict[str, Dict[str, Any]] = {}

    @abstractmethod
    def _call_api(self, response: str, **kwargs) -> Dict[str, Any]:
        """Make the actual API call. Override in subclasses."""
        pass

    def evaluate_response(self, response: str, **kwargs) -> Dict[str, Any]:
        """Evaluate response with caching and retries."""
        import hashlib
        import time

        cache_key = hashlib.md5(response.encode()).hexdigest()
        if self.cache_responses and cache_key in self._cache:
            return self._cache[cache_key]

        result = self._call_api(response, **kwargs)
        if self.cache_responses:
            self._cache[cache_key] = result
        return result

    def evaluate_batch(self, responses: List[str], **kwargs) -> List[Dict[str, Any]]:
        """Evaluate batch with optional batched API call."""
        if hasattr(self, '_call_api_batch'):
            return self._call_api_batch(responses, **kwargs)
        return super().evaluate_batch(responses, **kwargs)



"""NURT (Current) steering method constants and search space bounds."""
from wisent.core.utils.config_tools.constants.for_experiments.by_method.transport._szlak import *  # noqa: F401,F403

# --- NURT code-enforced bounds ---
SS_NURT_VAR_THRESH_LOW = 0.0
SS_NURT_VAR_THRESH_HIGH = 1.0
SS_NURT_INTEGRATION_STEPS_MIN = 1
SS_NURT_INTEGRATION_STEPS_MAX = 50
SS_NURT_T_MAX_LOW = -1.0
SS_NURT_T_MAX_HIGH = 1.0
SS_NURT_FLOW_HIDDEN_DIM_MIN = 0
SS_NURT_FLOW_HIDDEN_DIM_MAX = 256



"""Perplexity evaluation for LM-Eval-Harness Ground Truth."""

import logging
import numpy as np
import torch
from typing import Any, Dict

from wisent.core.primitives.model_interface.core.activations.activations import Activations
from wisent.core.utils.config_tools.constants import CONTEXT_MAX_PREVIEW, DISPLAY_TRUNCATION_MEDIUM, DISPLAY_TOP_N_SMALL, SPLIT_RATIO_FULL, DEFAULT_RANDOM_SEED
from wisent.core.primitives.models.core.layer import Layer
from wisent.core.primitives.models import get_generate_kwargs

logger = logging.getLogger(__name__)


def calculate_perplexity(model, text: str) -> float:
    """Calculate perplexity of text using the model."""
    try:
        prepared = model.prepare_activations(text)
        outputs, inputs = prepared["outputs"], prepared["inputs"]
        input_ids = inputs["input_ids"]
        logits = outputs.logits
        log_probs = torch.log_softmax(logits, dim=-1)
        if input_ids.shape[1] > 1:
            target_ids = input_ids[0, 1:]
            prediction_logits = log_probs[0, :-1, :]
            token_log_probs = prediction_logits.gather(dim=-1, index=target_ids.unsqueeze(-1)).squeeze(-1)
            avg_log_prob = token_log_probs.mean().item()
            perplexity = np.exp(-avg_log_prob)
        else:
            perplexity = float("inf")
        return perplexity
    except Exception as e:
        logger.error(f"Error calculating perplexity: {e}")
        return float("inf")


def evaluate_perplexity(evaluator, classifier, task_name: str, num_samples: int, model, layer: int,
                         token_aggregation: str) -> Dict[str, Any]:
    """Evaluate classifier using perplexity approach."""
    try:
        logger.info(f"PERPLEXITY EVALUATION: {task_name}")
        task_data = model.load_lm_eval_task(task_name, shots=0, limit=num_samples)
        docs, _ = model.split_task_data(task_data, split_ratio=SPLIT_RATIO_FULL, random_seed=DEFAULT_RANDOM_SEED)
        if not docs:
            return evaluator._error_result(f"No documents retrieved from task: {task_name}")
        logger.info(f"Retrieved {len(docs)} documents from {task_name}")
        perplexity_results = []
        for i, doc in enumerate(docs):
            try:
                if task_name == "wikitext":
                    text = doc.get("page", doc.get("text", ""))
                    if not text:
                        continue
                    perplexity = calculate_perplexity(model, text)
                    classification_score = None
                    try:
                        layer_obj = Layer(index=layer, type="transformer")
                        activation_text = text[:CONTEXT_MAX_PREVIEW] if len(text) > CONTEXT_MAX_PREVIEW else text
                        activation_tensor = model.extract_activations(activation_text, layer_obj)
                        activation_method = evaluator._map_token_aggregation_to_activation_method(token_aggregation)
                        activation_obj = Activations(tensor=activation_tensor, layer=layer_obj, aggregation_strategy=activation_method)
                        if classifier is not None:
                            features = activation_obj.extract_features_for_classifier()
                            try:
                                prediction_proba = classifier.predict_proba([features.cpu().numpy()])
                                if isinstance(prediction_proba, (list, tuple)) and len(prediction_proba) > 0:
                                    classification_score = float(prediction_proba[0])
                                else:
                                    classification_score = float(prediction_proba)
                                if hasattr(classification_score, "__len__") and not isinstance(classification_score, str):
                                    classification_score = float(classification_score[0])
                            except Exception:
                                predictions = classifier.predict([features.cpu().numpy()])
                                classification_score = float(predictions[0]) if len(predictions) > 0 else None
                        else:
                            classification_score = None
                    except Exception as e:
                        logger.error(f"Error classifying WikiText document: {e}")
                        classification_score = None
                    perplexity_results.append({"document_idx": i, "text_preview": text[:DISPLAY_TRUNCATION_MEDIUM] + "..." if len(text) > DISPLAY_TRUNCATION_MEDIUM else text,
                        "text_length": len(text), "perplexity": perplexity, "classifier_score": classification_score})
                    continue
                if hasattr(task_data, "doc_to_text"):
                    prompt = task_data.doc_to_text(doc)
                else:
                    prompt = str(doc.get("question", doc.get("text", "")))
                choices = []
                if hasattr(task_data, "doc_to_choice"):
                    choices = [task_data.doc_to_choice(doc, choice_idx) for choice_idx in range(len(doc.get("choices", [])))]
                elif "choices" in doc:
                    choices = doc["choices"]
                else:
                    gen_kwargs = get_generate_kwargs(do_sample=False)
                    generated_response, _ = model.generate(prompt=prompt, layer_index=layer, **gen_kwargs)
                    choices = [generated_response]
                choice_perplexities = []
                for choice_idx, choice in enumerate(choices):
                    try:
                        full_text = f"{prompt} {choice}"
                        perplexity = calculate_perplexity(model, full_text)
                        choice_perplexities.append({"choice_idx": choice_idx, "choice_text": choice, "perplexity": perplexity})
                    except Exception as e:
                        logger.error(f"Error calculating perplexity for choice {choice_idx}: {e}")
                        continue
                ground_truth_idx = None
                if hasattr(task_data, "doc_to_target"):
                    ground_truth = task_data.doc_to_target(doc)
                    try:
                        ground_truth_idx = int(ground_truth)
                    except:
                        ground_truth_idx = None
                elif "answer" in doc:
                    ground_truth_idx = doc["answer"]
                if choice_perplexities:
                    best_choice = min(choice_perplexities, key=lambda x: x["perplexity"])
                    classification_score = None
                    try:
                        layer_obj = Layer(index=layer, type="transformer")
                        activation_tensor = model.extract_activations(best_choice["choice_text"], layer_obj)
                        activation_method = evaluator._map_token_aggregation_to_activation_method(token_aggregation)
                        activation_obj = Activations(tensor=activation_tensor, layer=layer_obj, aggregation_strategy=activation_method)
                        features = activation_obj.extract_features_for_classifier()
                        try:
                            prediction_proba = classifier.predict_proba([features.cpu().numpy()])
                            if isinstance(prediction_proba, (list, tuple)) and len(prediction_proba) > 0:
                                classification_score = prediction_proba[0]
                            else:
                                classification_score = prediction_proba
                            if hasattr(classification_score, "__len__") and not isinstance(classification_score, str):
                                classification_score = classification_score[0] if len(classification_score) > 0 else None
                            classification_score = float(classification_score)
                        except Exception:
                            predictions = classifier.predict([features.cpu().numpy()])
                            classification_score = float(predictions[0]) if len(predictions) > 0 else None
                    except Exception as e:
                        logger.error(f"Error classifying best choice: {e}")
                    perplexity_results.append({"question": prompt, "choices": choice_perplexities,
                        "best_choice_idx": best_choice["choice_idx"], "best_choice_text": best_choice["choice_text"],
                        "best_choice_perplexity": best_choice["perplexity"], "ground_truth_idx": ground_truth_idx,
                        "classifier_score": classification_score,
                        "perplexity_correct": best_choice["choice_idx"] == ground_truth_idx if ground_truth_idx is not None else None})
            except Exception as e:
                logger.error(f"Error processing doc {i}: {e}")
                continue
        total_samples = len(perplexity_results)
        if task_name == "wikitext":
            perplexities = [r["perplexity"] for r in perplexity_results if r["perplexity"] != float("inf")]
            avg_perplexity = sum(perplexities) / len(perplexities) if perplexities else float("inf")
            classifier_scores = [r["classifier_score"] for r in perplexity_results if r["classifier_score"] is not None]
            avg_classifier_score = sum(classifier_scores) / len(classifier_scores) if classifier_scores else None
            perplexity_accuracy = 1.0 if avg_perplexity < 100 else 0.0
            correct_perplexity = sum(1 for r in perplexity_results if r["perplexity"] < 100)
        else:
            correct_perplexity = sum(1 for r in perplexity_results if r.get("perplexity_correct") == True)
            perplexity_accuracy = correct_perplexity / total_samples if total_samples > 0 else 0.0
            classifier_scores = [r["classifier_score"] for r in perplexity_results if r["classifier_score"] is not None]
            avg_classifier_score = sum(classifier_scores) / len(classifier_scores) if classifier_scores else None
        result_dict = {"ground_truth": "EVALUATED", "method_used": "lm-eval-harness-perplexity",
            "confidence": perplexity_accuracy, "details": f"Calculated perplexity for {total_samples} samples",
            "task_name": task_name, "evaluation_method": "perplexity", "perplexity_accuracy": perplexity_accuracy,
            "average_classifier_score": avg_classifier_score, "total_samples": total_samples,
            "correct_perplexity": correct_perplexity, "perplexity_results": perplexity_results[:DISPLAY_TOP_N_SMALL]}
        if task_name == "wikitext":
            result_dict["average_perplexity"] = avg_perplexity
            result_dict["details"] = f"Calculated perplexity for {total_samples} WikiText documents, avg perplexity: {avg_perplexity:.3f}"
        return result_dict
    except Exception as e:
        logger.error(f"Error in perplexity evaluation: {e}")
        return evaluator._error_result(f"Perplexity evaluation error: {e!s}")
